import json
from typing import List, Tuple, Dict
import os

import torch

from relogic.logickit.dataflow import DataFlow, Example, Feature, MiniBatch
from relogic.logickit.tokenizer.tokenizer_roberta_xlm import RobertaXLMTokenizer
from relogic.logickit.utils import create_tensor, filter_head_prediction
from transformers.tokenization_utils import PreTrainedTokenizer
from relogic.logickit.tokenizer.fasttext_tokenization import FasttextTokenizer


MAX_SEQ_LENGTH=500

class SequenceExample(Example):
  """SequenceExample contains the attributes and functionality of an Sequence Labeling example.

  Args:
    text (str): A sentence string.
    labels (List[str]): A list of labels.

  History Log: Originally we use adapted tokenizer for determining the token head
      for the pretokenized sentence. Here we plan to change it with for loop.
      Then we do not need to adapt it for each tokenizer.
      ```
      self.text_tokens, self.text_is_head = tokenizer.tokenize(self.text)
      ```
      We will adapt the code from transformer example run_ner.py and utils_ner.py
  """
  def __init__(self, text, labels=None, lang=None):
    super(SequenceExample, self).__init__()
    self.text = text
    self.raw_tokens = text.split(" ")
    self.labels = labels
    self.label_padding = "_X_"
    self.lang = lang
    self.padding_id = 0

  def process(self, tokenizers: Dict, *inputs, **kwargs):
    """Process the Sequence Example..
    """
    for tokenizer in tokenizers.values():
      if isinstance(tokenizer, PreTrainedTokenizer):
        # BERT process part
        self.text_tokens = []
        self.text_is_head = []
        for word in self.raw_tokens:
          word_tokens = tokenizer.tokenize(word)
          if len(word_tokens) == 0:
            self.text_tokens.append("[UNK]")
            self.text_is_head.append(1)
          else:
            self.text_tokens.extend(word_tokens)
            self.text_is_head.extend([1] + [0] * (len(word_tokens) - 1))
        if len(self.text_tokens) > MAX_SEQ_LENGTH:
          self.text_tokens = self.text_tokens[:MAX_SEQ_LENGTH]
          self.text_is_head = self.text_is_head[:MAX_SEQ_LENGTH]

        self.tokens = ["[CLS]"] + self.text_tokens + ["[SEP]"]
        self.segment_ids = [0] * (len(self.tokens))
        self.is_head = [2] + self.text_is_head + [2]
        self.head_index = [idx for idx, value in enumerate(self.is_head) if value == 1]

        self.input_ids = tokenizer.convert_tokens_to_ids(self.tokens)
        self.input_mask = [1] * len(self.input_ids)

        if self.labels is not None:
          label_mapping = kwargs.get("label_mapping")
          self.label_padding_id = label_mapping[self.label_padding]
          self.label_ids = [self.label_padding_id] * len(self.input_ids)
          for idx, label in zip(self.head_index, self.labels):
            self.label_ids[idx] = label_mapping[label]
        else:
          self.label_ids = None

        if self.lang is not None:
          language_name2id = kwargs.get("language_name2id")
          if language_name2id is not None:
            self.lang_id = language_name2id[self.lang]
      if isinstance(tokenizer, RobertaXLMTokenizer):
        # RobertaXLM process part
        (self.tokens, self.is_head,
         self.input_ids) = tokenizer.tokenize_and_add_placeholder_and_convert_to_ids(self.text, self.raw_tokens)
        self.segment_ids = [0] * len(self.tokens)
        self.head_index = [idx for idx, value in enumerate(self.is_head) if value == 1]

        self.input_mask = [1] * len(self.input_ids)

        if self.labels is not None:
          label_mapping = kwargs.get("label_mapping")
          self.label_padding_id = label_mapping[self.label_padding]
          self.label_ids = [self.label_padding_id] * len(self.input_ids)
          for idx, label in zip(self.head_index, self.labels):
            self.label_ids[idx] = label_mapping[label]
        else:
          self.label_ids = None

        self.padding_id = 1
      if isinstance(tokenizer, FasttextTokenizer):
        self._text_tokens = tokenizer.tokenize(self.text)
        self._input_token_ids = tokenizer.convert_tokens_to_ids(self._text_tokens)
        self._input_token_mask = [True] * len(self._text_tokens)
        if self.labels is not None:
          label_mapping = kwargs.get("label_mapping")
          self.label_padding_id = label_mapping[self.label_padding]
          self._label_ids = [label_mapping[label] for label in self.labels]
        else:
          self._label_ids = None

  @classmethod
  def from_structure(cls, structure):
    return cls(text=structure.tokenized_text)

  @classmethod
  def from_json(cls, example):
    return cls(text=" ".join(example["tokens"]),
               labels=example.get("labels", None),
               lang=example.get("lang", None))

  @property
  def len(self):
    return len(self.input_ids)

  @property
  def _len(self):
    return len(self._input_token_ids)

class SequenceFeature(Feature):
  """Sequence Features.
  """
  def __init__(self, *inputs, **kwargs):
    super(SequenceFeature, self).__init__()
    self.input_ids = kwargs.pop("input_ids")
    self.input_mask = kwargs.pop("input_mask")
    self.segment_ids = kwargs.pop("segment_ids")
    self.is_head = kwargs.pop("is_head")
    self.label_ids = kwargs.pop("label_ids")
    self.lang_ids = kwargs.pop("lang_ids", None)

    # Classical Feature
    self._input_token_ids = kwargs.pop("_input_token_ids")
    self._token_length = kwargs.pop("_token_length")
    self._label_ids = kwargs.pop("_label_ids")
    self._input_token_mask = kwargs.pop("_input_token_mask")

class SequenceMiniBatch(MiniBatch):
  """

  """
  def __init__(self, *inputs, **kwargs):
    super(SequenceMiniBatch, self).__init__(*inputs, **kwargs)

  def generate_input(self, device, use_label):
    inputs = {}
    inputs["task_name"] = self.task_name
    inputs["input_ids"] = create_tensor(self.input_features, "input_ids",
                                        torch.long, device)
    inputs["input_mask"] = create_tensor(self.input_features, "input_mask",
                                         torch.long, device)
    inputs["segment_ids"] = create_tensor(self.input_features, "segment_ids",
                                          torch.long, device)
    inputs["input_head"] = create_tensor(self.input_features, "is_head",
                                         torch.long, device)
    if use_label:
      label_ids = create_tensor(self.input_features, "label_ids",
                                torch.long, device)
      inputs["label_ids"] = label_ids
    else:
      inputs["label_ids"] = None

    inputs["lang_ids"] = create_tensor(self.input_features, "lang_ids",
                                       torch.long, device)

    inputs["extra_args"] = {}
    if self.config.tasks[self.task_name]["selected_non_final_layers"] is not None:
      inputs["extra_args"]["selected_non_final_layers"] = self.config.tasks[self.task_name]["selected_non_final_layers"]

    # Classical Features
    inputs["_input_token_ids"] = create_tensor(self.input_features, "_input_token_ids",
                                               torch.long, device)
    inputs["_token_length"] = create_tensor(self.input_features, "_token_length",
                                            torch.long, device)

    # if inputs["_input_token_ids"] is not None:
    #   batch_size = inputs["_input_token_ids"].size(0)
    #   sequence_length = inputs["_input_token_ids"].size(1)
    #
    #   _tmp = torch.arange(0, sequence_length).long().expand(batch_size, sequence_length).to(device)
    #   inputs["_input_token_mask"] = _tmp < inputs["_token_length"].unsqueeze(1).expand_as(_tmp).contiguous()
    #   del _tmp
    # else:
    #   inputs["_input_token_mask"] = None
    inputs["_input_token_mask"] = create_tensor(self.input_features, "_input_token_mask",
                                                torch.bool, device)

    if use_label:
      inputs["_label_ids"] = create_tensor(self.input_features, "_label_ids",
                                           torch.long, device)
    else:
      inputs["_label_ids"] = None

    return inputs

class SequenceDataFlow(DataFlow):
  def __init__(self, config, task_name, tokenizers, label_mapping):
    super(SequenceDataFlow, self).__init__(config, task_name, tokenizers, label_mapping)
    self._inv_label_mapping = {v: k for k, v in label_mapping.items()}

  @property
  def example_class(self):
    return SequenceExample

  @property
  def minibatch_class(self):
    return SequenceMiniBatch

  def process_example(self, example: SequenceExample):
    example.process(tokenizers=self.tokenizers,
                    label_mapping=self.label_mapping,
                    language_name2id=None)
                    # self.config.language_name2id)
    # We defer the language_name2id process to the base config process

  def convert_examples_to_features(self, examples: List[SequenceExample]):
    examples: List[SequenceExample]
    features = []

    try:
      max_token_length = max([example.len for example in examples])
    except:
      max_token_length = None

    try:
      _max_token_length = max([example._len for example in examples])
    except:
      _max_token_length = None

    for idx, example in enumerate(examples):
      if max_token_length is not None:
        padding = [0] * (max_token_length - example.len)
        input_ids = example.input_ids + [example.padding_id] * (max_token_length - example.len)
        input_mask = example.input_mask + padding
        segment_ids = example.segment_ids + padding
        is_head = example.is_head + [2] * (max_token_length - example.len)

        if example.label_ids is not None:
          # We assume the label length is same as sequence length
          label_ids = example.label_ids + [example.label_padding_id] * (max_token_length - example.len)
        else:
          label_ids = None

        if example.lang is not None:
          # These code is for XLM
          lang = [example.lang_id] * max_token_length
        else:
          lang = None
      else:
        input_ids, input_mask, is_head, segment_ids, label_ids, lang = None, None, None, None, None, None

      if _max_token_length is not None:
        _input_token_ids = example._input_token_ids + [0] * (_max_token_length - example._len)
        _token_length = example._len
        _input_token_mask = example._input_token_mask + [False] * (_max_token_length - example._len)
        if example._label_ids is not None:
          _label_ids = example._label_ids + [example.label_padding_id] * (_max_token_length - example._len)
        else:
          _label_ids = None
      else:
        _input_token_ids, _token_length, _input_token_mask, _label_ids = None, None, None, None
      features.append(SequenceFeature(
        input_ids=input_ids,
        input_mask=input_mask,
        is_head=is_head,
        segment_ids=segment_ids,
        label_ids=label_ids,
        lang=lang,
        _input_token_ids=_input_token_ids,
        _token_length=_token_length,
        _input_token_mask=_input_token_mask,
        _label_ids=_label_ids))

    return features

  def decode_to_labels(self, preds, mb: MiniBatch):
    labels = []
    for example, pred_logits in zip(mb.examples, preds[mb.task_name]["logits"]):
      pred_label_index = pred_logits.argmax(-1).data.cpu().numpy()
      pred_label = [self._inv_label_mapping[y_pred] for y_pred in pred_label_index]
      is_head = example.is_head if hasattr(example, "is_head") else example._input_token_mask
      pred_label = filter_head_prediction(sentence_tags=pred_label, is_head=is_head)
      labels.append(pred_label)
    return labels






from dataclasses import dataclass
from typing import List

from relogic.structures.document import Document
from relogic.structures.sentence import Sentence


@dataclass
class ReaderDataStructure:
  doc_list: List[Document]
  query: Sentence




from relogic.logickit.scorer.scorer import Scorer
import numpy as np
class DepParsingScorer(Scorer):
  def __init__(self, label_mapping, dump_to_file=None):
    super().__init__()
    self.labeled_correct = 0.0
    self.unlabeled_correct = 0.0
    self.exact_labeled_correct = 0.0
    self.exact_unlabeled_correct = 0.0
    self.total_words = 0.0
    self.total_sentences = 0.0

  def _get_results(self):
    """
    Returns
    -------
    The accumulated metrics as a dictionary.
    """
    unlabeled_attachment_score = 0.0
    labeled_attachment_score = 0.0
    unlabeled_exact_match = 0.0
    labeled_exact_match = 0.0
    if self.total_words > 0.0:
      unlabeled_attachment_score = float(self.unlabeled_correct) / float(self.total_words)
      labeled_attachment_score = float(self.labeled_correct) / float(self.total_words)
    if self.total_sentences > 0:
      unlabeled_exact_match = float(self.exact_unlabeled_correct) / float(
        self.total_sentences
      )
      labeled_exact_match = float(self.exact_labeled_correct) / float(self.total_sentences)
    return [
      ("total_sents", self.total_sentences),
      ("total_words", self.total_words),
      ("UAS", unlabeled_attachment_score * 100.0),
      ("LAS", labeled_attachment_score * 100.0),
      ("UEM", unlabeled_exact_match * 100.0),
      ("LEM", labeled_exact_match * 100.0),
    ]


  def update(self, mb, predictions, loss, extra_args):
    predicted_indices = predictions["heads"]
    predicted_labels = predictions["head_tags"]
    # We know there will be mask here
    masks = predictions["mask"]

    predicted_indices, predicted_labels, masks = self.unwrap_to_tensors(predicted_indices, predicted_labels, masks)
    for each_predicted_indices, each_predicted_labels, example, mask in zip(predicted_indices, predicted_labels, mb.examples, masks):
      gold_indices = np.array(example.arcs_ids[1:-1])[np.array(example.is_head[1:-1]) == 1]
      gold_label_ids = np.array(example.label_ids[1:-1])[np.array(example.is_head[1:-1]) == 1]
      each_predicted_indices = each_predicted_indices[1:][mask[1:] == 1].numpy()
      each_predicted_labels = each_predicted_labels[1:][mask[1:] == 1].numpy()

      assert(len(gold_indices) == len(each_predicted_indices))

      correct_indices = (gold_indices == each_predicted_indices)
      correct_labels = (gold_label_ids == each_predicted_labels)
      correct_labels_and_indices = np.logical_and(correct_indices, correct_labels)
      exact_unlabeled_correct = (correct_indices.sum() == len(correct_indices))
      exact_labeled_correct = (correct_labels_and_indices.sum() == len(correct_indices))

      self.unlabeled_correct += correct_indices.sum()
      self.exact_unlabeled_correct += exact_unlabeled_correct
      self.labeled_correct += correct_labels_and_indices.sum()
      self.exact_labeled_correct += exact_labeled_correct
      self.total_sentences += 1
      self.total_words += len(correct_indices)

  def get_loss(self):
    return 0






import torch


class InputVariationalDropout(torch.nn.Dropout):
  """
  Apply the dropout technique in Gal and Ghahramani, "Dropout as a Bayesian Approximation:
  Representing Model Uncertainty in Deep Learning" (https://arxiv.org/abs/1506.02142) to a
  3D tensor.
  This module accepts a 3D tensor of shape ``(batch_size, num_timesteps, embedding_dim)``
  and samples a single dropout mask of shape ``(batch_size, embedding_dim)`` and applies
  it to every time step.
  """
  def forward(self, input_tensor):
    """
    Apply dropout to input tensor.
    Parameters
    ----------
    input_tensor: ``torch.FloatTensor``
        A tensor of shape ``(batch_size, num_timesteps, embedding_dim)``
    Returns
    -------
    output: ``torch.FloatTensor``
        A tensor of shape ``(batch_size, num_timesteps, embedding_dim)`` with dropout applied.
    """
    ones = input_tensor.data.new_ones(input_tensor.shape[0],
                                      input_tensor.shape[-1])
    dropout_mask = torch.nn.functional.dropout(ones,
                                               self.p,
                                               self.training,
                                               inplace=False)
    if self.inplace:
      input_tensor *= dropout_mask.unsqueeze(1)
      return None
    else:
      return dropout_mask.unsqueeze(1) * input_tensor



import torch.nn as nn
import torch
from relogic.logickit.modules.span_extractors import EndpointSpanExtractor, AttentiveSpanExtractor 


class ECPExtractionModule(nn.Module):
  """Emotion-Cause Extraction
  """
  def __init__(self, config, task_name, n_classes):
    super(ECPExtractionModule, self).__init__()
    self.config = config
    self.task_name = task_name
    self.n_classes = n_classes

    self.endpoint_span_extractor = EndpointSpanExtractor(
      input_dim=config.hidden_size,
      num_width_embeddings=config.num_width_embeddings,
      span_width_embedding_dim=config.span_width_embedding_dim)
    
    hidden_size = config.hidden_size

    self.attentive_span_extractor = AttentiveSpanExtractor(input_dim=hidden_size)
  
  def forward(self, *inputs, **kwargs):
    features = kwargs.pop("features")
    clause_candidates = kwargs.pop("clause_candidates")

    clause_mask = (clause_candidates[:, :, 1] > 0).float()

    endpoint_span_embeddings = self.endpoint_span_extractor(
      sequence_tensor=features, span_indices=clause_candidates, span_indices_mask=clause_mask)

    raise NotImplementedError("Play with it")
    

    
    



import argparse
import os


def main(input_folder, output_folder):
  for fold in ["train"]:
    path = os.path.join(input_folder, fold+".txt")
    target_path = os.path.join(output_folder, fold+".txt")
    left_to_right = {}
    right_to_left = {}
    left = {}
    right = {}
    with open(path) as fin:
      for line in fin:
        if line.startswith("Left-Right-Query"):
          continue
        id_pair, comm_split1, comm_split2, label, gold_id_pair = line.strip().split('\t')
        split1, split2, direc = id_pair.split('-')
        if direc == "0":
          if split1 not in left_to_right:
            left_to_right[split1] = [None, []]
          if label == '1':
            left_to_right[split1][0] = split2
          else:
            left_to_right[split1][1].append(split2)
          left[split1] = comm_split1
          right[split2] = comm_split2
        if direc == "1":
          if split2 not in right_to_left:
            right_to_left[split2] = [None, []]
          if label == '1':
            right_to_left[split2][0] = split1
          else:
            right_to_left[split2][1].append(split1)
          left[split1] = comm_split1
          right[split2] = comm_split2
    with open(target_path, 'w') as fout:
      for split1 in left_to_right.keys():
        text_a = left[split1] # query
        gold_can = left_to_right[split1][0] # correct candidate
        text_b = right[gold_can]
        for neg in left_to_right[split1][1]:
          pair_id = "{}-{}-0".format(split1, neg)
          gold_id = "{}-{}-0".format(split1, gold_can)
          text_c = right[neg]
          fout.write("{}\t{}\t{}\t{}\t{}\n".format(
            pair_id, text_a, text_b, text_c, gold_id))
      for split2 in right_to_left.keys():
        text_a = right[split2]
        gold_can = right_to_left[split2][0]
        text_b = left[gold_can]
        for neg in right_to_left[split2][1]:
          pair_id = "{}-{}-0".format(neg, split2)
          gold_id = "{}-{}-0".format(gold_can, split2)
          text_c = left[neg]
          fout.write("{}\t{}\t{}\t{}\t{}\n".format(
            pair_id, text_a, text_b, text_c, gold_id))
  for fold in ["dev", "test"]:
    path = os.path.join(input_folder, fold + ".txt")
    target_path = os.path.join(output_folder, fold + ".txt")
    left = {}
    right = {}
    left_to_right_gold = {}
    right_to_left_gold = {}
    with open(path) as fin:
      for line in fin:
        if line.startswith("Left-Right-Query"):
          continue
        id_pair, comm_split1, comm_split2, label, gold_id_pair = line.strip().split('\t')
        split1, split2, direc = id_pair.split('-')
        if direc == "0":
          left[split1] = comm_split1
          right[split2] = comm_split2
        else:
          left[split1] = comm_split1
          right[split2] = comm_split2
        split1, split2, direc = gold_id_pair.split('-')
        if direc == "0":
          left_to_right_gold[split1] = split2
        else:
          right_to_left_gold[split2] = split1
    with open(target_path, "w") as fout:
      for split1 in left_to_right_gold:
        guid = split1 + "-none-0"
        gold_id = split1 + "-{}-0".format(left_to_right_gold[split1])
        fout.write("{}\t{}\t{}\n".format(
          guid, left[split1], gold_id))
      for split2 in right_to_left_gold:
        guid = "none-" + split2 + "-1"
        gold_id = right_to_left_gold[split2] + "-{}-1".format(split2)
        fout.write("{}\t{}\t{}\n".format(
          guid, right[split2], gold_id))



if __name__ == "__main__":
  parser = argparse.ArgumentParser()
  parser.add_argument("--input_folder")
  parser.add_argument("--output_folder")
  args = parser.parse_args()
  if not os.path.exists(args.output_folder):
    os.makedirs(args.output_folder)
  main(args.input_folder, args.output_folder)


""" Official evaluation script for v1.1 of the SQuAD dataset. """
from __future__ import print_function
from collections import Counter
import string
import re
import argparse
import json
import sys


def normalize_answer(s):
  """Lower text and remove punctuation, articles and extra whitespace."""

  def remove_articles(text):
    return re.sub(r'\b(a|an|the)\b', ' ', text)

  def white_space_fix(text):
    return ' '.join(text.split())

  def remove_punc(text):
    exclude = set(string.punctuation)
    return ''.join(ch for ch in text if ch not in exclude)

  def lower(text):
    return text.lower()

  return white_space_fix(remove_articles(remove_punc(lower(s))))


def f1_score(prediction, ground_truth):
  prediction_tokens = normalize_answer(prediction).split()
  ground_truth_tokens = normalize_answer(ground_truth).split()
  common = Counter(prediction_tokens) & Counter(ground_truth_tokens)
  num_same = sum(common.values())
  if num_same == 0:
    return 0
  precision = 1.0 * num_same / len(prediction_tokens)
  recall = 1.0 * num_same / len(ground_truth_tokens)
  f1 = (2 * precision * recall) / (precision + recall)
  return f1


def exact_match_score(prediction, ground_truth):
  return (normalize_answer(prediction) == normalize_answer(ground_truth))


def metric_max_over_ground_truths(metric_fn, prediction, ground_truths):
  scores_for_ground_truths = []
  for ground_truth in ground_truths:
    score = metric_fn(prediction, ground_truth)
    scores_for_ground_truths.append(score)
  return max(scores_for_ground_truths)


def evaluate(dataset, predictions, na_prob_thresh=None):
  f1 = exact_match = total = 0
  for article in dataset:
    for paragraph in article['paragraphs']:
      for qa in paragraph['qas']:
        total += 1
        if qa['id'] not in predictions:
          message = 'Unanswered question ' + qa['id'] + \
                    ' will receive score 0.'
          print(message, file=sys.stderr)
          continue
        ground_truths = list(map(lambda x: x['text'], qa['answers']))
        prediction = predictions[qa['id']]
        exact_match += metric_max_over_ground_truths(exact_match_score,
                                                     prediction, ground_truths)
        f1 += metric_max_over_ground_truths(f1_score, prediction,
                                            ground_truths)

  exact_match = 100.0 * exact_match / total
  f1 = 100.0 * f1 / total

  return [('exact_match', exact_match), ('f1', f1)]


if __name__ == '__main__':
  expected_version = '1.1'
  parser = argparse.ArgumentParser(
    description='Evaluation for SQuAD ' + expected_version)
  parser.add_argument('dataset_file', help='Dataset file')
  parser.add_argument('prediction_file', help='Prediction File')
  args = parser.parse_args()
  with open(args.dataset_file) as dataset_file:
    dataset_json = json.load(dataset_file)
    if (dataset_json['version'] != expected_version):
      print(
        'Evaluation expects v-' + expected_version +
        ', but got dataset with v-' + dataset_json['version'],
        file=sys.stderr)
    dataset = dataset_json['data']
  with open(args.prediction_file) as prediction_file:
    predictions = json.load(prediction_file)
  print(json.dumps(evaluate(dataset, predictions)))



"""
1. Read raw data
2. Preprocess:
    convert to id
    save length
    padding sequence to max length
    bucket them
3. Save to hd5f
4. Load file with max_memory_limit
5. Shuffle
6. Slice to remove extra paddings
"""

class LazyDataset(object):
  def __init__(self):
    pass


from __future__ import absolute_import, division, print_function

import argparse
import json
import os
import random
from types import SimpleNamespace

import numpy as np

import torch
from relogic.logickit.base import utils
from relogic.logickit.base.configure import configure, update_configure
from relogic.logickit.training import trainer, training_progress
from relogic.logickit.serving import Server
from relogic.logickit.analyzer.heads_importance import compute_heads_importance, mask_heads

if "PUDB" not in os.environ or os.environ["PUDB"] == "false":
  import relogic.utils.crash_on_ipy


def train(config):
  if config.use_external_teacher:
    teacher_model_path = config.teacher_model_path
    teacher_config = os.path.join(teacher_model_path, "general_config.json")
    with open(teacher_config) as f:
      teacher_config = SimpleNamespace(**json.load(f))
    teacher_config.local_rank = config.local_rank
    teacher_config.no_cuda = config.no_cuda
  else:
    teacher_config = None
  # A quick fix for loading external teacher
  model_trainer = trainer.Trainer(
    config=config, teacher_config=teacher_config)
  # A quick fix for version migration
  progress = training_progress.TrainingProgress(config=config)
  if config.use_external_teacher:
    model_path = os.path.join(teacher_model_path,
                              teacher_config.model_name + ".ckpt")
    model_trainer.restore(model_path)
    model_trainer.restore_teacher(model_path)
  model_trainer.train(progress)

def finetune(config):
  general_config_path = os.path.join(config.finetune_restore_path,
                                     "general_config.json")
  with open(general_config_path) as f:
    restore_config = SimpleNamespace(**json.load(f))
  if config.model_name:
    model_path = os.path.join(config.finetune_restore_path,
                            config.model_name + ".ckpt")
  else:
    model_path = os.path.join(config.finetune_restore_path,
                              restore_config.model_name + ".ckpt")

  model_trainer = trainer.Trainer(config)
  model_trainer.restore(model_path)
  progress = training_progress.TrainingProgress(config=config)
  model_trainer.train(progress)


def eval(config):
  general_config_path = os.path.join(config.restore_path,
                                     "general_config.json")
  with open(general_config_path) as f:
    restore_config = SimpleNamespace(**json.load(f))
  if config.model_name:
    model_path = os.path.join(config.restore_path,
                            config.model_name + ".ckpt")
  else:
    model_path = os.path.join(config.restore_path,
                              restore_config.model_name + ".ckpt")
  restore_config.mode = config.mode
  restore_config.output_features = config.output_features
  restore_config.local_rank = config.local_rank
  restore_config.no_cuda = config.no_cuda
  restore_config.buckets = config.buckets
  restore_config.gold_answer_file = config.gold_answer_file
  restore_config.null_score_diff_threshold = config.null_score_diff_threshold
  restore_config.output_attentions = config.output_attentions
  restore_config.use_external_teacher = False
  if not hasattr(restore_config, "branching_encoder"):
    restore_config.branching_encoder = False
  # Update the evaluation dataset
  update_configure(restore_config, config)
  print(restore_config)
  utils.heading("RUN {} ({:})".format(config.mode.upper(),
                                      restore_config.task_names))
  model_trainer = trainer.Trainer(restore_config)
  model_trainer.restore(model_path)
  if config.mode == "serving":
    server = Server(model_trainer)
    server.start()
  elif config.mode == "analysis":
    analyze(config, model_trainer)
  elif config.mode == "feature_extraction":
    task_ = None
    for task in model_trainer.tasks:
      if task.name == config.selected_task:
        task_ = task
    model_trainer.feature_extraction(
      task=task_, dump_file=config.feature_dump_file)
  else:
    model_trainer.evaluate_all_tasks()

def analyze(config, model_trainer):
  # compute_heads_importance(config, model_trainer)
  mask_heads(config, model_trainer)




def main():
  utils.heading("SETUP")
  parser = argparse.ArgumentParser()

  # IO
  parser.add_argument(
    "--mode", default=None, choices=["train", "valid", "eval", "finetune", "analysis", "feature_extraction"])
  parser.add_argument("--output_dir", type=str, default="data/models")
  parser.add_argument("--max_seq_length", type=int, default=450)
  parser.add_argument("--max_query_length", type=int, default=64)
  parser.add_argument("--doc_stride", type=int, default=128)
  parser.add_argument("--do_lower_case", default=False, action="store_true")
  parser.add_argument("--model_name", type=str)
  parser.add_argument("--restore_path", type=str)
  parser.add_argument("--finetune_restore_path", type=str)
  parser.add_argument("--train_file", type=str, default="train.json")
  parser.add_argument("--dev_file", type=str, default="dev.json")
  parser.add_argument("--test_file", type=str, default="test.json")

  # Task Definition
  parser.add_argument("--task_names", type=str)
  parser.add_argument("--raw_data_path", type=str)
  parser.add_argument("--label_mapping_path", type=str)
  parser.add_argument("--unsupervised_data", type=str)
  parser.add_argument("--lang", type=str, default="en")
  parser.add_argument("--pretokenized", action="store_true", default=False)
  parser.add_argument("--topk", default=1)
  parser.add_argument("--gold_answer_file", default="data/preprocessed_data/squad20.json")
  parser.add_argument("--dump_to_files_dict", default="")

  parser.add_argument("--output_attentions", default=False, action="store_true")
  parser.add_argument("--span_inference", default=False, action="store_true")
  parser.add_argument("--metrics", default="", type=str)

  # Task related configuration

  # Sequence Labeling
  parser.add_argument("--sequence_labeling_use_cls", default=False, action="store_true")

  # Relation Extraction
  parser.add_argument("--no_entity_surface", dest="entity_surface_aware", default=True, action="store_false")
  parser.add_argument("--use_dependency_feature", dest="use_dependency_feature", default=False, action="store_true")
  parser.add_argument("--rel_extraction_module_type", type=str, default="hybrid")

  # Semantic Role Labeling
  parser.add_argument("--no_predicate_surface", dest="predicate_surface_aware", default=True, action="store_false")
  parser.add_argument("--no_span_annotation", dest="use_span_annotation", default=True, action="store_false")
  parser.add_argument("--use_span_candidates", default=False, action="store_true")
  parser.add_argument("--srl_module_type", type=str, default="sequence_labeling")
  parser.add_argument("--label_embed_dim", type=int, default=100)
  parser.add_argument("--external_vocab_embed_dim", type=int, default=300)
  parser.add_argument("--external_embeddings", type=str)
  parser.add_argument("--use_description", default=False, action="store_true")
  parser.add_argument("--srl_label_format", default="srl_label_span_based", type=str)
  parser.add_argument("--num_width_embeddings", type=int, default=300)
  parser.add_argument("--span_width_embedding_dim", type=int, default=100)
  parser.add_argument("--srl_candidate_loss", default=False, action="store_true")
  parser.add_argument("--srl_arg_span_repr", default="ave")
  parser.add_argument("--srl_pred_span_repr", default="ave")
  parser.add_argument("--srl_use_label_embedding", default=False, action="store_true")
  parser.add_argument("--srl_compute_pos_tag_loss", default=False, action="store_true")
  parser.add_argument("--srl_use_gold_predicate", default=False, action="store_true")
  parser.add_argument("--srl_use_gold_argument", default=False, action="store_true")
  parser.add_argument("--predicate_reveal_method", default=None, type=str)
  parser.add_argument("--indicator_embedding_size", default=10, type=int)

  # Dependency Parsing
  parser.add_argument("--dep_parsing_mlp_dim", default=300, type=int)
  parser.add_argument("--dropout", default=0.3, type=float)

  # Parallel Mapping
  parser.add_argument("--parallel_mapping_mode", default="alignment", type=str)

  # Reading Comprehension
  parser.add_argument("--null_score_diff_threshold", default=1.0)

  # Information Retrieval
  parser.add_argument("--qrels_file_path", type=str, default=None)
  parser.add_argument("--regression", default=False, action="store_true")
  parser.add_argument("--word_level_interaction", default=False, action="store_true")
  parser.add_argument("--ir_siamese", default=False, action="store_true")

  # CNN model
  parser.add_argument("--output_channel", type=int, default=150)
  parser.add_argument("--kernel_size", type=int, default=2)
  parser.add_argument("--word_embed_dim", type=int, default=300)

  # Modeling
  parser.add_argument("--use_gcn", dest="use_gcn", default=False, action="store_true")
  parser.add_argument("--fix_embedding", default=False, action="store_true")

  # Model
  parser.add_argument("--bert_model", type=str)
  parser.add_argument("--encoder_type", type=str, default="bert", choices=["bert", "xlm", "xlmr", "lstm", "embedding"])
  parser.add_argument("--hidden_size", type=int, default=768)
  parser.add_argument("--projection_size", type=int, default=300)
  parser.add_argument(
    "--initializer_range", type=float,
    default=0.02)  # initialization for task module
  # follow the initialization range of bert
  parser.add_argument("--no_bilstm", default=True, dest="use_bilstm", action="store_false")
  parser.add_argument("--repr_size", default=300, type=int)
  parser.add_argument("--branching_encoder", default=False, action="store_true")
  parser.add_argument("--routing_config_file", type=str)
  parser.add_argument("--selected_non_final_layers", type=str, default="none", help="split by ; among tasks")
  parser.add_argument("--dataset_type", type=str, default="bucket")
  parser.add_argument("--language_id_file", type=str, default=None)

  # Semi-Supervised
  parser.add_argument("--is_semisup", default=False, action="store_true")
  parser.add_argument("--partial_view_sources", type=str)
  parser.add_argument("--use_external_teacher", default=False, action="store_true")
  parser.add_argument("--teacher_model_path", default=None, type=str)

  # Training
  parser.add_argument("--seed", type=int, default=3435)
  parser.add_argument("--no_cuda", action="store_true")
  parser.add_argument("--local_rank", type=int, default=-1)
  parser.add_argument("--learning_rate", type=float, default=5e-5)
  parser.add_argument("--warmup_proportion", type=float, default=0.1)
  parser.add_argument(
    "--gradient_accumulation_steps",
    type=int,
    default=1,
    help=
    "Number of updates steps to accumulate before performing a backward/update pass"
  )
  parser.add_argument("--print_every", type=int, default=25)
  parser.add_argument("--eval_dev_every", default=2000, type=int)
  parser.add_argument("--train_batch_size", type=str, default="8")
  parser.add_argument("--test_batch_size", type=str, default="8")
  parser.add_argument("--grad_clip", type=float, default=1.0)
  parser.add_argument("--epoch_number", type=int, default=20)
  parser.add_argument("--self_attention_head_size", default=64, type=int)
  parser.add_argument("--schedule_method", default="warmup_linear")
  parser.add_argument(
    "--no_schedule_lr", dest="schedule_lr", default=True, action="store_false")
  parser.add_argument("--word_dropout", default=False, action="store_true")
  parser.add_argument("--word_dropout_prob", default=0.6, type=float)
  parser.add_argument("--max_margin", type=float, default=3)
  parser.add_argument("--warmup_epoch_number", type=int, default=0)
  parser.add_argument("--sgd_learning_rate", type=float, default=0.1)
  parser.add_argument("--adam_learning_rate", type=float, default=0.003)
  parser.add_argument("--sep_optim", dest="sep_optim", default=False, action="store_true")
  parser.add_argument("--multi_gpu", dest="multi_gpu", default=False, action="store_true")
  parser.add_argument("--ignore_parameters", default="", type=str)
  parser.add_argument("--fix_bert", default=False, action="store_true")
  parser.add_argument("--two_stage_optim", default=False, action="store_true")
  parser.add_argument("--training_scheme", default=None, type=str)
  parser.add_argument("--training_scheme_file", default=None, type=str)
  parser.add_argument("--num_train_optimization_steps", default=0, type=int)
  parser.add_argument("--early_stop_at", default=0, type=int)
  parser.add_argument("--loss_weight", type=str, default='1')
  parser.add_argument("--select_index_method", type=str, default="cls")
  parser.add_argument("--use_cosine_loss", default=False, action="store_true")
  parser.add_argument("--adversarial_training", default=None, type=str)
  parser.add_argument("--no_bucket", default=False, action="store_true")
  parser.add_argument("--param_initialization", default=None, type=str)
  # We allow to set same training steps for different dataset
  # Need to combine to CUDA_VISIBLE_DEVICES
  parser.add_argument("--only_adam", default=False, action="store_true")

  # Analysis
  parser.add_argument("--head_to_mask_file", type=str, default="")


  # Configuration
  parser.add_argument("--config_file", type=str, default=None)
  parser.add_argument("--trainer_config", type=str, default=None)
  parser.add_argument("--module_config", type=str, default=None)
  parser.add_argument("--task_config", type=str, default=None)

  #
  parser.add_argument("--selected_task", type=str)
  parser.add_argument("--feature_dump_file", type=str)
  parser.add_argument("--output_features", default=False, action="store_true")

  args = parser.parse_args()

  if not args.mode:
    raise ValueError("You need to specify the mode")
  if args.output_dir:
    if os.path.exists(args.output_dir) and os.listdir(
        args.output_dir) and args.mode == "train":
      raise ValueError(
        "Output directory ({}) already exists and is not empty.".format(
          args.output_dir))
    if not os.path.exists(args.output_dir):
      os.makedirs(args.output_dir)

  if args.gradient_accumulation_steps < 1:
    raise ValueError(
      "Invalid gradient_accumulation_steps parameter: {}, should be >= 1".
      format(args.gradient_accumulation_steps))
  # args.train_batch_size = args.train_batch_size // args.gradient_accumulation_steps
  # num_train_optimization_steps = len(train_examples) / batch_size * epoch_number

  random.seed(args.seed)
  np.random.seed(args.seed)
  torch.manual_seed(args.seed)
  torch.cuda.manual_seed_all(args.seed)

  configure(args)

  print(args)

  if args.mode == "train":
    utils.heading("START TRAINING ({:})".format(args.task_names))
    train(args)
  elif args.mode == "valid":
    eval(args)
  elif args.mode == "eval":
    eval(args)
  elif args.mode == "finetune":
    finetune(args)
  elif args.mode == "serving":
    eval(args)
  elif args.mode == "analysis":
    eval(args)
  elif args.mode == "feature_extraction":
    eval(args)


if __name__ == "__main__":
  main()



from typing import List
from relogic.structures.structure import Structure
from relogic.utils.file_utils import cached_path, RELOGIC_CACHE
import logging
import os
import json
from types import SimpleNamespace
from relogic.logickit.training.predictor import Predictor
from relogic.logickit.base.configuration import Argument




logger = logging.getLogger(__name__)

PRETRAINED_MODEL_ARCHIVE_MAP = {
  "ner-zh": "",
  "ner-en": "",
  "entity-linking": "",
  "srl-conll12": "https://git.uwaterloo.ca/p8shi/data-server/raw/master/srl_conll12_mlp/default.ckpt",
  "pd-conll12": "https://git.uwaterloo.ca/p8shi/data-server/raw/master/pd_conll12_lstm/default.ckpt"
}

PRETRAINED_CONFIG_ARCHIVE_MAP = {
  "ner-zh": "",
  "ner-en": "",
  "entity-linking": "",
  "srl-conll12": "https://git.uwaterloo.ca/p8shi/data-server/raw/master/srl_conll12_mlp/general_config.json",
  "pd-conll12": "https://git.uwaterloo.ca/p8shi/data-server/raw/master/pd_conll12_lstm/general_config.json"
}

WEIGHTS_NAME = "default.ckpt"
CONFIG_NAME = "general_config.json"

class Component(object):
  """

  """
  def __init__(self, model_name, config, predictor: Predictor=None):
    self.model_name = model_name
    self.config = config
    self._predictor = predictor

  def execute(self, inputs: List[Structure]):
    raise NotImplementedError()

  @classmethod
  def from_pretrained(cls, pretrained_model_name_or_path, *inputs, **kwargs):
    if pretrained_model_name_or_path in PRETRAINED_MODEL_ARCHIVE_MAP:
      archive_file = PRETRAINED_MODEL_ARCHIVE_MAP[pretrained_model_name_or_path]
      config_file = PRETRAINED_CONFIG_ARCHIVE_MAP[pretrained_model_name_or_path]
    else:
      archive_file = os.path.join(pretrained_model_name_or_path, WEIGHTS_NAME)
      config_file = os.path.join(pretrained_model_name_or_path, CONFIG_NAME)

    # redirect to the cache, if necessary
    try:
      resolved_archive_file = cached_path(archive_file, cache_dir=RELOGIC_CACHE)
    except EnvironmentError:
      if pretrained_model_name_or_path in PRETRAINED_MODEL_ARCHIVE_MAP:
        logger.error(
          "Couldn't reach server at '{}' to download pretrained weights.".format(
            archive_file))
      else:
        logger.error(
          "Model name '{}' was not found in model name list ({}). "
          "We assumed '{}' was a path or url but couldn't find any file "
          "associated to this path or url.".format(
            pretrained_model_name_or_path,
            ', '.join(PRETRAINED_MODEL_ARCHIVE_MAP.keys()),
            archive_file))
      return None

    try:
      resolved_config_file = cached_path(config_file, cache_dir=RELOGIC_CACHE)
    except EnvironmentError:
      if pretrained_model_name_or_path in PRETRAINED_CONFIG_ARCHIVE_MAP:
        logger.error(
          "Couldn't reach server at '{}' to download pretrained model configuration file.".format(
            config_file))
      else:
        logger.error(
          "Model name '{}' was not found in model name list ({}). "
          "We assumed '{}' was a path or url but couldn't find any file "
          "associated to this path or url.".format(
            pretrained_model_name_or_path,
            ', '.join(PRETRAINED_CONFIG_ARCHIVE_MAP.keys()),
            config_file))
      return None

    if resolved_archive_file == archive_file and resolved_config_file == config_file:
      logger.info("loading weights file {}".format(archive_file))
      logger.info("loading configuration file {}".format(config_file))
    else:
      logger.info("loading weights file {} from cache at {}".format(
        archive_file, resolved_archive_file))
      logger.info("loading configuration file {} from cache at {}".format(
        config_file, resolved_config_file))


    resolved_config_file_dir = os.path.dirname(resolved_config_file)
    resolved_config_file_name = os.path.basename(resolved_config_file)
    restore_config = Argument.restore_from_model_path(model_path=resolved_config_file_dir, config_name=resolved_config_file_name)
    predictor = Predictor(restore_config)
    resolved_model_file_dir = os.path.dirname(resolved_archive_file)
    resolved_model_file_name = os.path.basename(resolved_archive_file)
    predictor.restore(model_path=resolved_model_file_dir, model_name=resolved_model_file_name)

    # The model name here is not that reasonable, sometimes it will be a self defined path.
    return cls(model_name=pretrained_model_name_or_path, config=restore_config, predictor=predictor)


import argparse
import io
import os
import collections
from tqdm import tqdm
import random
import numpy as np

def load_vectors(fname):
  fin = io.open(fname, 'r', encoding='utf-8', newline='\n', errors='ignore')
  try:
    n, d = map(int, fin.readline().split())
  except:
    print("No heading")
    d = int(input("You need to specify the dim of the vector"))
  data = []
  vocab= ["<PAD>", "<UNK>"]
  # data.append([random.uniform(-0.1, 0.1) for i in range(d)])
  # data.append([random.uniform(-0.1, 0.1) for i in range(d)])
  data.append([0 for i in range(d)])
  data.append([0 for i in range(d)])
  index = 2
  for line in tqdm(fin):
    tokens = line.rstrip().split(' ')
    data.append(list(map(float, tokens[1:])))
    vocab.append(tokens[0])
    index += 1

  return vocab, data

def main(vec_file_path, vocab_file_path, vector_file_path):
  vocab, data = load_vectors(vec_file_path)
  with open(vocab_file_path, 'w') as fout:
    for token in vocab:
      fout.write(token + "\n")
  np.save(vector_file_path, np.asarray(data))


if __name__ == "__main__":
  parser = argparse.ArgumentParser()
  parser.add_argument("--vec_file_path", type=str)
  parser.add_argument("--vocab_file_path", type=str)
  parser.add_argument("--vector_file_path", type=str)
  args = parser.parse_args()
  main(args.vec_file_path, args.vocab_file_path, args.vector_file_path)


import tensorflow as tf
from relogic.tpukit.inference import modeling
from relogic.tpukit.model import optimization


def create_model(bert_config, is_training, input_ids, input_mask, segment_ids,
                 labels, num_labels, use_one_hot_embeddings):
  """Creates a classification model."""
  model = modeling.BertModel(
      config=bert_config,
      is_training=is_training,
      input_ids=input_ids,
      input_mask=input_mask,
      token_type_ids=segment_ids,
      use_one_hot_embeddings=use_one_hot_embeddings)

  # In the demo, we are doing a simple classification task on the entire
  # segment.
  #
  # If you want to use the token-level output, use model.get_sequence_output()
  # instead.
  output_layer = model.get_pooled_output()

  hidden_size = output_layer.shape[-1].value

  output_weights = tf.get_variable(
      "output_weights", [num_labels, hidden_size],
      initializer=tf.truncated_normal_initializer(stddev=0.02))

  output_bias = tf.get_variable(
      "output_bias", [num_labels], initializer=tf.zeros_initializer())

  with tf.variable_scope("loss"):
    if is_training:
      # I.e., 0.1 dropout
      output_layer = tf.nn.dropout(output_layer, keep_prob=0.9)

    logits = tf.matmul(output_layer, output_weights, transpose_b=True)
    logits = tf.nn.bias_add(logits, output_bias)
    probabilities = tf.nn.softmax(logits, axis=-1)
    log_probs = tf.nn.log_softmax(logits, axis=-1)

    one_hot_labels = tf.one_hot(labels, depth=num_labels, dtype=tf.float32)

    per_example_loss = -tf.reduce_sum(one_hot_labels * log_probs, axis=-1)
    loss = tf.reduce_mean(per_example_loss)

    return (loss, per_example_loss, logits, probabilities)

def model_fn_builder(bert_config, num_labels, init_checkpoint, learning_rate,
                     num_train_steps, num_warmup_steps, use_tpu,
                     use_one_hot_embeddings):
  """Returns `model_fn` closure for TPUEstimator."""

  def model_fn(features, labels, mode, params):  # pylint: disable=unused-argument
    """The `model_fn` for TPUEstimator."""

    tf.logging.info("*** Features ***")
    for name in sorted(features.keys()):
      tf.logging.info("  name = %s, shape = %s" % (name, features[name].shape))

    input_ids = features["input_ids"]
    input_mask = features["input_mask"]
    segment_ids = features["segment_ids"]
    label_ids = features["label_ids"]
    is_real_example = None
    if "is_real_example" in features:
      is_real_example = tf.cast(features["is_real_example"], dtype=tf.float32)
    else:
      is_real_example = tf.ones(tf.shape(label_ids), dtype=tf.float32)

    is_training = (mode == tf.estimator.ModeKeys.TRAIN)

    (total_loss, per_example_loss, logits, probabilities) = create_model(
        bert_config, is_training, input_ids, input_mask, segment_ids, label_ids,
        num_labels, use_one_hot_embeddings)

    tvars = tf.trainable_variables()
    initialized_variable_names = {}
    scaffold_fn = None
    if init_checkpoint:
      (assignment_map, initialized_variable_names
      ) = modeling.get_assignment_map_from_checkpoint(tvars, init_checkpoint)
      if use_tpu:

        def tpu_scaffold():
          tf.train.init_from_checkpoint(init_checkpoint, assignment_map)
          return tf.train.Scaffold()

        scaffold_fn = tpu_scaffold
      else:
        tf.train.init_from_checkpoint(init_checkpoint, assignment_map)

    tf.logging.info("**** Trainable Variables ****")
    for var in tvars:
      init_string = ""
      if var.name in initialized_variable_names:
        init_string = ", *INIT_FROM_CKPT*"
      tf.logging.info("  name = %s, shape = %s%s", var.name, var.shape,
                      init_string)

    output_spec = None
    if mode == tf.estimator.ModeKeys.TRAIN:

      train_op = optimization.create_optimizer(
          total_loss, learning_rate, num_train_steps, num_warmup_steps, use_tpu)

      output_spec = tf.contrib.tpu.TPUEstimatorSpec(
          mode=mode,
          loss=total_loss,
          train_op=train_op,
          scaffold_fn=scaffold_fn)
    elif mode == tf.estimator.ModeKeys.EVAL:

      def metric_fn(per_example_loss, label_ids, logits, is_real_example):
        predictions = tf.argmax(logits, axis=-1, output_type=tf.int32)
        accuracy = tf.metrics.accuracy(
            labels=label_ids, predictions=predictions, weights=is_real_example)
        loss = tf.metrics.mean(values=per_example_loss, weights=is_real_example)
        return {
            "eval_accuracy": accuracy,
            "eval_loss": loss,
        }

      eval_metrics = (metric_fn,
                      [per_example_loss, label_ids, logits, is_real_example])
      output_spec = tf.contrib.tpu.TPUEstimatorSpec(
          mode=mode,
          loss=total_loss,
          eval_metrics=eval_metrics,
          scaffold_fn=scaffold_fn)
    else:
      output_spec = tf.contrib.tpu.TPUEstimatorSpec(
          mode=mode,
          predictions={"probabilities": probabilities},
          scaffold_fn=scaffold_fn)
    return output_spec

  return model_fn


import torch
import torch.nn as nn


class SpanExtractionModule(nn.Module):
  def __init__(self, config, task_name, n_classes):
    super(SpanExtractionModule, self).__init__()
    self.config = config
    self.task_name = task_name
    self.n_classes = n_classes
    self.to_logits = nn.Linear(config.hidden_size, self.n_classes)
    self.apply(self.init_weights)

  def init_weights(self, module):
    if isinstance(module, nn.Linear):
      # Slightly different from the TF version which uses truncated_normal for initialization
      # cf https://github.com/pytorch/pytorch/pull/5617
      module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
    if isinstance(module, nn.Linear) and module.bias is not None:
      module.bias.data.zero_()


  def forward(self, input_feature, input_mask=None, segment_ids=None, extra_args=None, **kwargs):
    logits = self.to_logits(input_feature)
    # (batch_size, sentence_length, 2)
    start_logits, end_logits = logits.split(1, dim=-1)
    start_logits = start_logits.squeeze(-1)
    end_logits = end_logits.squeeze(-1)

    return start_logits, end_logits





import os
import json
from relogic.logickit.base.constants import NEVER_SPLIT

def configure(config):
  config.buckets = [(0, 40), (40, 80), (80, config.max_seq_length)]
  if config.task_names and config.restore_path is None:
    config.task_names = config.task_names.split(',')
    config.raw_data_path = config.raw_data_path.split(',')
    config.label_mapping_path = config.label_mapping_path.split(',')
    config.train_file = config.train_file.split(',')
    config.dev_file = config.dev_file.split(',')
    config.test_file = config.test_file.split(',')
    config.loss_weight = config.loss_weight.split(',')
    config.selected_non_final_layers = config.selected_non_final_layers.split(';')
    config.dataset_type = config.dataset_type.split(',')
    if config.qrels_file_path is not None:
      config.qrels_file_path = config.qrels_file_path.split(',')

    # I assume those tasks do not use qrels will put none as placeholder
    config.train_batch_size = config.train_batch_size.split(',')
    config.test_batch_size = config.test_batch_size.split(',')

    if len(config.train_file) != len(config.task_names):
      config.train_file = [config.train_file[0]] * len(config.task_names)
      config.dev_file = [config.dev_file[0]] * len(config.task_names)
      config.test_file = [config.test_file[0]] * len(config.task_names)
    if len(config.loss_weight) != len(config.task_names):
      config.loss_weight = [config.loss_weight[0]] * len(config.task_names)
    if len(config.selected_non_final_layers) != len(config.task_names):
      config.selected_non_final_layers = [config.selected_non_final_layers[0]] * len(config.task_names)
    if len(config.dataset_type) != len(config.task_names):
      config.dataset_type = [config.dataset_type[0]] * len(config.task_names)

    if len(config.train_batch_size) != len(config.task_names):
      config.train_batch_size = [config.train_batch_size[0]] * len(config.task_names)
    if len(config.test_batch_size) != len(config.task_names):
      config.test_batch_size = [config.test_batch_size[0]] * len(config.task_names)

    assert len(config.task_names) == len(config.raw_data_path) == len(config.label_mapping_path)
    config.tasks = {}
    for (task, raw_data_path, label_mapping_path, train_file, dev_file, test_file,
         loss_weight, selected_non_final_layers, dataset_type, train_batch_size, test_batch_size) in zip(
          config.task_names, config.raw_data_path, config.label_mapping_path,
          config.train_file, config.dev_file, config.test_file, config.loss_weight,
          config.selected_non_final_layers, config.dataset_type,
      config.train_batch_size, config.test_batch_size):
      config.tasks[task] = {}
      config.tasks[task]["raw_data_path"] = raw_data_path
      config.tasks[task]["label_mapping_path"] = label_mapping_path
      config.tasks[task]["train_file"] = train_file
      config.tasks[task]["dev_file"] = dev_file
      config.tasks[task]["test_file"] = test_file
      config.tasks[task]["loss_weight"] = float(loss_weight)
      config.tasks[task]["selected_non_final_layers"] = None if selected_non_final_layers == "none" else [
        int(item) for item in selected_non_final_layers.split(',')]
      config.tasks[task]["dataset_type"] = dataset_type
      config.tasks[task]["train_batch_size"] = int(train_batch_size)
      config.tasks[task]["test_batch_size"] = int(test_batch_size)
    if config.qrels_file_path is not None:
      for (task, qrels_file_path) in zip(config.task_names, config.qrels_file_path):
        config.tasks[task]["qrels_file_path"] = qrels_file_path

  if config.output_dir:
    config.progress = os.path.join(config.output_dir, "progress")
    config.history_file = os.path.join(config.output_dir, "history.pkl")
  if config.partial_view_sources:
    config.partial_view_sources = [int(x) for x in config.partial_view_sources.split(',')]

  config.never_split = NEVER_SPLIT
  if config.branching_encoder:
    routing_config = json.load(open(config.routing_config_file))
    config.task_route_paths = routing_config["task_route_paths"]
    config.branching_structure = routing_config["branching_structure"]
  config.ignore_parameters = list(filter(lambda x:len(x.strip())>0, config.ignore_parameters.split(",")))
  config.vocab_path = config.bert_model
  if config.language_id_file is not None:
    config.language_name2id = json.load(config.language_id_file)
  else:
    config.language_name2id = None
  if config.task_names:
    if "rel_extraction" in config.task_names:
      if config.bert_model not in ["bert-base-cased", "bert-large-cased"]:
        raise ValueError("For relation extraction on tacred, the vocab only support bert-base-cased for masking")
      config.vocab_path = "relogic/logickit/vocabs/tacred-{}-vocab.txt".format(config.bert_model)

  config.external_vocab_size = 999996 # a quick patch
  config.external_vocab_embed_size = 300
  if config.metrics != "":
    config.metrics = config.metrics.split(',')
  else:
    config.metrics = None

def update_configure(restore_config, config):
  # quick fix for batch_size
  for task_name in restore_config.tasks:
    if "train_batch_size" not in restore_config.tasks[task_name]:
      restore_config.tasks[task_name]["train_batch_size"] = restore_config.train_batch_size
      restore_config.tasks[task_name]["test_batch_size"] = restore_config.test_batch_size

  if config.qrels_file_path is not None:
    restore_config.qrels_file_path = config.qrels_file_path
  if config.raw_data_path:
    assert config.task_names is not None
    # If user want to change the raw_data_path, they need to provide the 
    # task_name and the raw_data_path. These two arguments should match 
    # each other.
    config.raw_data_path = config.raw_data_path.split(",")
    task_names = config.task_names.split(",")
    # If user want to change the raw_data_path, then they need to change
    # for all tasks.
    # assert len(config.raw_data_path) == len(config.tasks)
    for name, raw_data_path in zip(task_names, config.raw_data_path):
      restore_config.tasks[name]["raw_data_path"] = raw_data_path
  if config.task_names is not None:

    config.test_file = config.test_file.split(",")
    task_names = config.task_names.split(",")
    if len(config.test_file) == len(task_names):
      print("=========")
      print("We assume you are going to update the test file "
            "because you specify the task_names and test_file at the same time")
      for name, test_file in zip(task_names, config.test_file):
        restore_config.tasks[name]["test_file"] = test_file
    



from typing import List, Tuple

import torch

from relogic.logickit.dataflow import DataFlow, Example, Feature, MiniBatch
from relogic.logickit.utils import create_tensor
from relogic.logickit.tokenizer import CustomizedBertTokenizer, RobertaXLMTokenizer
from transformers.tokenization_utils import PreTrainedTokenizer



class ParallelExample(Example):
  """Parallel Example contains the attributes and functionality of
  a parallel sentence pair. This class is mainly used for auxiliary
  task of multi-lingual model training. Parallel sentence (in different
  language) should be provided and token level alignment is needed.

  Args:
    text_a (str): sentence in source language
    text_b (str): sentence in target language
    alignment (Tuple[List]): token level alignment. We plan to take the format
      that tuple of list. There are two list in the tuple, first one is
      the index for source sentence, second one is the index for target
      sentence, which are parallel in the corresponding position.
      If you have the data format same as fast_align, you can use the scripts
      to convert into this format. The conversion rules are described in the script.
  """
  def __init__(self, text_a: str, text_b: str, alignment: Tuple[List]):
    super(ParallelExample, self).__init__()
    self.text_a = text_a
    self.text_b = text_b
    self.alignment = alignment
    self.padding_id = 0

  def process(self, tokenizers, *inputs, **kwargs):
    """Process the sentence pair."""

    for tokenizer in tokenizers.values():
      if isinstance(tokenizer, PreTrainedTokenizer):
        self.text_a_tokens = tokenizer.tokenize(self.text_a)
        self.text_b_tokens = tokenizer.tokenize(self.text_b)

        self.a_tokens = ["[CLS]"] + self.text_a_tokens + ["[SEP]"]
        self.a_segment_ids = [0] * len(self.a_tokens)

        self.b_tokens = ["[CLS]"] + self.text_b_tokens + ["[SEP]"]
        self.b_segment_ids = [0] * len(self.b_tokens)

        self.a_input_ids = tokenizer.convert_tokens_to_ids(self.a_tokens)
        self.b_input_ids = tokenizer.convert_tokens_to_ids(self.b_tokens)
        self.a_input_mask = [1] * len(self.a_input_ids)
        self.b_input_mask = [1] * len(self.b_input_ids)

        self.a_is_head = None
        self.b_is_head = None
        self.a_selected_indices = None
        self.b_selected_indices = None

      elif isinstance(tokenizer, CustomizedBertTokenizer):
        self.text_a_tokens, self.text_a_is_head = tokenizer.tokenize(self.text_a)
        self.text_b_tokens, self.text_b_is_head = tokenizer.tokenize(self.text_b)

        max_seq_length = kwargs.pop("max_seq_length", 512)
        self.text_a_tokens = self.text_a_tokens[:max_seq_length-2]
        self.text_b_tokens = self.text_b_tokens[:max_seq_length-2]
        # We need to keep position for [CLS] and [SEP]

        self.a_tokens = ["[CLS]"] + self.text_a_tokens + ["[SEP]"]
        self.a_segment_ids = [0] * len(self.a_tokens)
        self.a_is_head = [2] + self.text_a_is_head + [2]

        self.b_tokens = ["[CLS]"] + self.text_b_tokens + ["[SEP]"]
        self.b_segment_ids = [0] * len(self.b_tokens)
        self.b_is_head = [2] + self.text_b_is_head + [2]

        self.a_input_ids = tokenizer.convert_tokens_to_ids(self.a_tokens)
        self.b_input_ids = tokenizer.convert_tokens_to_ids(self.b_tokens)
        self.a_input_mask = [1] * len(self.a_input_ids)
        self.b_input_mask = [1] * len(self.b_input_ids)

        self.a_head_index = [
          idx for idx, value in enumerate(self.a_is_head) if value == 1
        ] + [len(self.a_is_head) - 1]
        self.b_head_index = [
          idx for idx, value in enumerate(self.b_is_head) if value == 1
        ] + [len(self.b_is_head) - 1]

        if self.alignment is not None:
          self.a_selected_indices = [self.a_head_index[index] for index in self.alignment[0]]
          self.b_selected_indices = [self.b_head_index[index] for index in self.alignment[1]]
        else:
          self.a_selected_indices = None
          self.b_selected_indices = None
      elif isinstance(tokenizer, RobertaXLMTokenizer):
        self.padding_id = 1
        (self.a_tokens, self.a_is_head,
         self.a_input_ids) = tokenizer.tokenize_and_add_placeholder_and_convert_to_ids(
          self.text_a, None)
        (self.b_tokens, self.b_is_head,
         self.b_input_ids) = tokenizer.tokenize_and_add_placeholder_and_convert_to_ids(
          self.text_b, None)
        self.a_segment_ids = [0] * len(self.a_tokens)
        self.b_segment_ids = [0] * len(self.b_tokens)
        self.a_input_mask = [1] * len(self.a_input_ids)
        self.b_input_mask = [1] * len(self.b_input_ids)

        # self.a_head_index = [
        #                       idx for idx, value in enumerate(self.a_is_head) if value == 1
        #                     ] + [len(self.a_is_head) - 1]
        # self.b_head_index = [
        #                       idx for idx, value in enumerate(self.b_is_head) if value == 1
        #                     ] + [len(self.b_is_head) - 1]
        #
        # if self.alignment is not None:
        #   self.a_selected_indices = [self.a_head_index[index] for index in self.alignment[0]]
        #   self.b_selected_indices = [self.b_head_index[index] for index in self.alignment[1]]
        # else:
        self.a_selected_indices = None
        self.b_selected_indices = None




  @classmethod
  def from_structure(cls, structure):
    return cls(text_a=structure.text_a,
               text_b=structure.text_b,
               alignment=structure.alignment)

  @classmethod
  def from_json(cls, example):
    return cls(text_a=example["text_a"],
               text_b=example["text_b"],
               alignment=example.get("alignment", None))

  @property
  def len(self):
    return len(self.a_tokens) + len(self.b_tokens)

  @property
  def len_a(self):
    return len(self.a_tokens)

  @property
  def len_b(self):
    return len(self.b_tokens)

class ParallelFeature(Feature):
  """Parallel Feature

  """
  def __init__(self, *input, **kwargs):
    super(ParallelFeature, self).__init__()
    self.a_input_ids = kwargs.pop("a_input_ids")
    self.b_input_ids = kwargs.pop("b_input_ids")
    self.a_segment_ids = kwargs.pop("a_segment_ids")
    self.b_segment_ids = kwargs.pop("b_segment_ids")
    self.a_input_mask = kwargs.pop("a_input_mask")
    self.b_input_mask = kwargs.pop("b_input_mask")
    self.a_selected_indices = kwargs.pop("a_selected_indices")
    self.b_selected_indices = kwargs.pop("b_selected_indices")
    self.a_is_head = kwargs.pop("a_is_head")
    self.b_is_head = kwargs.pop("b_is_head")

class ParallelMiniBatch(MiniBatch):
  def __init__(self, *inputs, **kwargs):
    super(ParallelMiniBatch, self).__init__(*inputs, **kwargs)

  def generate_input(self, device, use_label):
    """Generate tensors based on Parallel Feature"""
    # BERT based features
    inputs = {}
    inputs["task_name"] = self.task_name
    inputs["a_input_ids"] = create_tensor(self.input_features, "a_input_ids",
                                          torch.long, device)
    inputs["b_input_ids"] = create_tensor(self.input_features, "b_input_ids",
                                          torch.long, device)
    inputs["a_input_mask"] = create_tensor(self.input_features, "a_input_mask",
                                           torch.long, device)
    inputs["b_input_mask"] = create_tensor(self.input_features, "b_input_mask",
                                           torch.long, device)
    inputs["a_segment_ids"] = create_tensor(self.input_features, "a_segment_ids",
                                            torch.long, device)
    inputs["b_segment_ids"] = create_tensor(self.input_features, "b_segment_ids",
                                            torch.long, device)
    inputs["a_selected_indices"] = create_tensor(self.input_features, "a_selected_indices",
                                            torch.long, device)
    inputs["b_selected_indices"] = create_tensor(self.input_features, "b_selected_indices",
                                            torch.long, device)
    inputs["a_is_head"] = create_tensor(self.input_features, "a_is_head",
                                            torch.long, device)
    inputs["b_is_head"] = create_tensor(self.input_features, "b_is_head",
                                            torch.long, device)

    inputs["extra_args"] = {}
    if self.config.tasks[self.task_name]["selected_non_final_layers"] is not None:
      inputs["extra_args"]["selected_non_final_layers"] = self.config.tasks[self.task_name]["selected_non_final_layers"]


    return inputs

class ParallelDataFlow(DataFlow):
  """DataFlow implementation based on Parallel Task"""
  def __init__(self, config, task_name, tokenizers, label_mapping):
    super(ParallelDataFlow, self).__init__(config, task_name, tokenizers, label_mapping)

  @property
  def example_class(self):
    return ParallelExample

  @property
  def minibatch_class(self):
    return ParallelMiniBatch

  def process_example(self, example: ParallelExample):
    example.process(tokenizers=self.tokenizers)

  def convert_examples_to_features(self, examples: List[ParallelExample]):
    examples: List[ParallelExample]
    features = []
    a_max_token_length = max([example.len_a for example in examples])
    b_max_token_length = max([example.len_b for example in examples])

    try:
      max_selected_indices_length = max([len(example.a_selected_indices) for example in examples])
    except:
      max_selected_indices_length = 0

    for idx, example in enumerate(examples):
      a_padding = [0] * (a_max_token_length - example.len_a)
      b_padding = [0] * (b_max_token_length - example.len_b)
      a_input_ids = example.a_input_ids + [example.padding_id] * (a_max_token_length - example.len_a)
      b_input_ids = example.b_input_ids + [example.padding_id] * (b_max_token_length - example.len_b)
      a_segment_ids = example.a_segment_ids + a_padding
      b_segment_ids = example.b_segment_ids + b_padding
      a_input_mask = example.a_input_mask + a_padding
      b_input_mask = example.b_input_mask + b_padding
      if example.a_is_head is not None and example.b_is_head is not None:
        a_is_head = example.a_is_head + a_padding
        b_is_head = example.b_is_head + b_padding
      else:
        a_is_head = None
        b_is_head = None

      if max_selected_indices_length > 0:
        a_selected_indices = example.a_selected_indices + [0] * (
            max_selected_indices_length - len(example.a_selected_indices))
        b_selected_indices = example.b_selected_indices + [0] * (
            max_selected_indices_length - len(example.b_selected_indices))
      else:
        a_selected_indices = None
        b_selected_indices = None

      features.append(
        ParallelFeature(
          a_input_ids=a_input_ids,
          b_input_ids=b_input_ids,
          a_input_mask=a_input_mask,
          b_input_mask=b_input_mask,
          a_segment_ids=a_segment_ids,
          b_segment_ids=b_segment_ids,
          a_selected_indices=a_selected_indices,
          b_selected_indices=b_selected_indices,
          a_is_head=a_is_head,
          b_is_head=b_is_head))
    return features


import argparse
import json
from relogic.structures.document import Sentence
from tqdm import tqdm
from relogic.pipelines.core import Pipeline

parser = argparse.ArgumentParser()
parser.add_argument("--data_file_path", type=str)
parser.add_argument("--output_file_path", type=str)

args = parser.parse_args()

pipeline = Pipeline(
  component_names=["predicate_detection", "srl"],
  component_model_names= {"predicate_detection" : "pd-conll12" ,"srl": "srl-conll12"})

fout = open(args.output_file_path, 'w')
with open(args.data_file_path) as fin:
  sentences = []
  for line in tqdm(fin):
    if line == "\n":
      pipeline.execute(sentences)
      for sent in sentences:
        fout.write(json.dumps(sent.convert_to_json()) + "\n")
      fout.write("\n")
      sentences = []
    else:
      sentences.append(Sentence(text=line.strip("\n"), tokenizer="gpt2"))








import os
import json
from relogic.logickit.base import utils
from relogic.logickit.dataflow import DataFlow, TASK_TO_DATAFLOW_CLASS_MAP


class LabeledDataLoader(object):
  def __init__(self, config, name, tokenizer):
    self.config = config
    self.task_name = name
    self.raw_data_path = config.tasks[name]["raw_data_path"]
    self.label_mapping_path = config.tasks[name]["label_mapping_path"]
    self.file_names = {
      "train": config.tasks[name]["train_file"],
      "dev": config.tasks[name]["dev_file"],
      "test": config.tasks[name]["test_file"]
    }
    self.tokenizer = tokenizer
    # TODO: these code can not support multi-label set dataset.
    # Will fix this in the future.
    if self.label_mapping:
      self.n_classes = len(set(self.label_mapping.values()))
    else:
      self.n_classes = None
    self.extra_args = {}

  def get_dataset(self, split):
    if (split == 'train'
          and (os.path.exists(os.path.join(self.raw_data_path, "train_subset.txt")) or
               os.path.exists(os.path.join(self.raw_data_path, "train_subset.json")))):
      split = 'train_subset'
    dataflow: DataFlow = self.get_dataflow()
    file_path = os.path.join(self.raw_data_path, self.file_names[split])
    dataflow.update_with_file(file_path)
    self.dataflow = dataflow
    return dataflow

  def get_dataflow(self) -> DataFlow:
    return TASK_TO_DATAFLOW_CLASS_MAP[self.task_name](
      config=self.config, task_name=self.task_name ,tokenizers=self.tokenizer, label_mapping=self.label_mapping)

  @property
  def label_mapping(self):
    if self.label_mapping_path == "none":
      return {}
    if self.label_mapping_path.endswith(".pkl"):
      return utils.load_pickle(self.label_mapping_path)
    elif self.label_mapping_path.endswith(".json"):
      return json.load(open(self.label_mapping_path))
    else:
      raise ValueError(self.label_mapping_path)


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
from relogic.logickit.base import utils
import torch
import json

class TrainingProgress(object):
  """
  TODO: We are going to retire the Training Progress somehow.
  1. Remove the unlabeled data. The unlabeled data interface will be integrated into
     dataflow.

  """
  def __init__(self, config, restore_if_possible=True):
    self.config = config
    if restore_if_possible and os.path.exists(config.progress):
      history, current_file, current_line = utils.load_pickle(
        config.progress, memoized=False)
      self.history = history
      # self.unlabeled_data_reader =
    else:
      utils.log("No previous checkpoint found - starting from scratch")
      self.history = []
    self.evaluated_steps = set([0])
    self.log_steps = set([])
    # We do not want to evaluate in step 0

  def write(self):
    utils.write_pickle(
      (self.history, self.unlabeled_data_reader.current_file,
       self.unlabeled_data_reader.current_line),
      self.config.progress)

  def save_if_best_dev_model(self, model):
    # Why it is average score here
    # TODO: double check the results format
    best_avg_score = 0
    for i, results in enumerate(self.history):
      for result in results:
        if any("train" in metric for metric, value in result):
          continue
        if any("test" in metric for metric, value in result):
          continue
      total, count = 0, 0
      for result in results:
        for metric, value in result:
          if hasattr(self.config, "metrics") and self.config.metrics is not None:
            if metric in self.config.metrics:
              total += value
              count += 1
          elif "distance" in metric or "f1" in metric or "las" in metric or "accuracy" in metric or "recall_left" in metric or "recall_right" in metric or "map" in metric:
            total += value
            count += 1
      avg_score = total / count
      if avg_score >= best_avg_score:
        best_avg_score = avg_score
        if i == len(self.history) - 1:
          utils.log("New Score {}, New best model! Saving ...".format(best_avg_score))
          torch.save(model.state_dict(), os.path.join(self.config.output_dir, self.config.model_name + ".ckpt"))
          general_config_path = os.path.join(self.config.output_dir, "general_config.json")
          with open(general_config_path, "w") as fout:
            fout.write(json.dumps(vars(self.config)))
          # TODO: finish model saving

  def evaluated_in_step(self, step):
    return step in self.evaluated_steps

  def add_evaluated_step(self, step):
    self.evaluated_steps.add(step)

  def log_in_step(self, step):
    return step in self.log_steps

  def add_log_step(self, step):
    self.log_steps.add(step)


"""
CoNLL English Example:

1	But	but	but	CC	CC	_	_	33	33	DEP	DEP	_	_	_	_	_	_	_
2	while	while	while	IN	IN	_	_	33	33	ADV	ADV	_	_	_	_	AM-ADV	_	_
3	the	the	the	DT	DT	_	_	7	7	NMOD	NMOD	_	_	_	_	_	_	_
4	New	new	new	NNP	NNP	_	_	5	5	NAME	NAME	_	_	_	_	_	_	_
5	York	york	york	NNP	NNP	_	_	7	7	NAME	NAME	_	_	_	_	_	_	_
6	Stock	stock	stock	NNP	NNP	_	_	7	7	NAME	NAME	_	_	_	_	_	_	_
7	Exchange	exchange	exchange	NNP	NNP	_	_	8	8	SBJ	SBJ	_	_	A1	_	_	_	_
"""
import argparse

def read_conll(path):
  with open(path) as fin:
    examples = []
    sentence = []
    for line in fin:
      line = line.strip()
      if len(line) == 0:
        examples.append(sentence)
        sentence = []
      else:
        sentence.append(line.split('\t'))
    if len(sentence) > 0:
      examples.append(sentence)
    return examples

def label_process(label):
  if label == '_':
    return 'O'
  else:
    return label.split('.')[1]

def output_sense(examples, output_path):
  with open(output_path, 'w') as fout:
    for example in examples:
      if len(example[0]) <= 14:
        continue
      for items in example:
        fout.write("{} {}\n".format(items[1], label_process(items[13])))
      fout.write("\n")

if __name__ == "__main__":
  parser = argparse.ArgumentParser()
  parser.add_argument("--conll_path")
  parser.add_argument("--output_path")
  args = parser.parse_args()
  examples = read_conll(args.conll_path)
  output_sense(examples, args.output_path)


# coding=utf-8
# Copyright 2018 The Google AI Language Team Authors and The HuggingFace Inc. team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tokenization classes."""

from __future__ import absolute_import, division, print_function, unicode_literals

import collections
import logging
import os
import unicodedata
from io import open

from relogic.utils.file_utils import cached_path

logger = logging.getLogger(__name__)

PRETRAINED_VOCAB_ARCHIVE_MAP = {
    'bert-base-uncased': "https://s3.amazonaws.com/models.huggingface.co/bert/bert-base-uncased-vocab.txt",
    'bert-large-uncased': "https://s3.amazonaws.com/models.huggingface.co/bert/bert-large-uncased-vocab.txt",
    'bert-base-cased': "https://s3.amazonaws.com/models.huggingface.co/bert/bert-base-cased-vocab.txt",
    'bert-large-cased': "https://s3.amazonaws.com/models.huggingface.co/bert/bert-large-cased-vocab.txt",
    'bert-base-multilingual-uncased': "https://s3.amazonaws.com/models.huggingface.co/bert/bert-base-multilingual-uncased-vocab.txt",
    'bert-base-multilingual-cased': "https://s3.amazonaws.com/models.huggingface.co/bert/bert-base-multilingual-cased-vocab.txt",
    'bert-base-chinese': "https://s3.amazonaws.com/models.huggingface.co/bert/bert-base-chinese-vocab.txt",
    'bert-base-german-cased': "https://int-deepset-models-bert.s3.eu-central-1.amazonaws.com/pytorch/bert-base-german-cased-vocab.txt",
    'bert-large-uncased-whole-word-masking': "https://s3.amazonaws.com/models.huggingface.co/bert/bert-large-uncased-whole-word-masking-vocab.txt",
    'bert-large-cased-whole-word-masking': "https://s3.amazonaws.com/models.huggingface.co/bert/bert-large-cased-whole-word-masking-vocab.txt",
    'bert-large-uncased-whole-word-masking-finetuned-squad': "https://s3.amazonaws.com/models.huggingface.co/bert/bert-large-uncased-whole-word-masking-finetuned-squad-vocab.txt",
    'bert-large-cased-whole-word-masking-finetuned-squad': "https://s3.amazonaws.com/models.huggingface.co/bert/bert-large-cased-whole-word-masking-finetuned-squad-vocab.txt",
    'bert-base-cased-finetuned-mrpc': "https://s3.amazonaws.com/models.huggingface.co/bert/bert-base-cased-finetuned-mrpc-vocab.txt",
}
PRETRAINED_VOCAB_POSITIONAL_EMBEDDINGS_SIZE_MAP = {
    'bert-base-uncased': 512,
    'bert-large-uncased': 512,
    'bert-base-cased': 512,
    'bert-large-cased': 512,
    'bert-base-multilingual-uncased': 512,
    'bert-base-multilingual-cased': 512,
    'bert-base-chinese': 512,
    'bert-base-german-cased': 512,
    'bert-large-uncased-whole-word-masking': 512,
    'bert-large-cased-whole-word-masking': 512,
    'bert-large-uncased-whole-word-masking-finetuned-squad': 512,
    'bert-large-cased-whole-word-masking-finetuned-squad': 512,
    'bert-base-cased-finetuned-mrpc': 512,
}
VOCAB_NAME = 'vocab.txt'


def load_vocab(vocab_file):
  """Loads a vocabulary file into a dictionary."""
  vocab = collections.OrderedDict()
  index = 0
  with open(vocab_file, "r", encoding="utf-8") as reader:
    while True:
      token = reader.readline()
      if not token:
        break
      token = token.strip()
      vocab[token] = index
      index += 1
  return vocab


def whitespace_tokenize(text):
  """Runs basic whitespace cleaning and splitting on a piece of text."""
  text = text.strip()
  if not text:
    return []
  tokens = text.split()
  return tokens


class BertTokenizer(object):
  """Runs end-to-end tokenization: punctuation splitting + wordpiece"""

  def __init__(self, vocab_file, do_lower_case=True, max_len=None, do_basic_tokenize=True,
               never_split=("[UNK]", "[SEP]", "[PAD]", "[CLS]", "[MASK]"), lang="", pretokenized=False):
    """Constructs a BertTokenizer.

    Args:
      vocab_file: Path to a one-wordpiece-per-line vocabulary file
      do_lower_case: Whether to lower case the input
                     Only has an effect when do_wordpiece_only=False
      do_basic_tokenize: Whether to do basic tokenization before wordpiece.
      max_len: An artificial maximum length to truncate tokenized sequences to;
                     Effective maximum length is always the minimum of this
                     value (if specified) and the underlying BERT model's
                     sequence length.
      never_split: List of tokens which will never be split during tokenization.
                     Only has an effect when do_wordpiece_only=False
    """
    if not os.path.isfile(vocab_file):
      raise ValueError(
        "Can't find a vocabulary file at path '{}'. To load the vocabulary from a Google pretrained "
        "model use `tokenizer = BertTokenizer.from_pretrained(PRETRAINED_MODEL_NAME)`".format(vocab_file))
    self.vocab = load_vocab(vocab_file)
    self.ids_to_tokens = collections.OrderedDict(
      [(ids, tok) for tok, ids in self.vocab.items()])
    self.do_basic_tokenize = do_basic_tokenize
    self.lang = lang
    if do_basic_tokenize:
      self.basic_tokenizer = BasicTokenizer(do_lower_case=do_lower_case,
                                            never_split=never_split,
                                            lang=lang, pretokenized=pretokenized)
    self.wordpiece_tokenizer = WordpieceTokenizer(vocab=self.vocab)
    self.max_len = max_len if max_len is not None else int(1e12)

  def tokenize(self, text):
    is_head = None
    if self.do_basic_tokenize:
      """
      To support the token level task, we add is_head indicator
        to denotes the head of the word piece.
      Currently this function does not support for directly using
        word piece tokenizer.
      """
      split_tokens = []
      is_head = []
      tokens, basic_tokenizer_is_head = self.basic_tokenizer.tokenize(text)
      for token, head in zip(tokens, basic_tokenizer_is_head):
        for sub_idx, sub_token in enumerate(self.wordpiece_tokenizer.tokenize(token)):
          split_tokens.append(sub_token)
          if sub_idx == 0 and head == 1:
            is_head.append(1)
          else:
            is_head.append(0)
    else:
      split_tokens = self.wordpiece_tokenizer.tokenize(text)
    return split_tokens, is_head

  def convert_tokens_to_ids(self, tokens):
    """Converts a sequence of tokens into ids using the vocab."""
    ids = []
    for token in tokens:
      ids.append(self.vocab[token])
    if len(ids) > self.max_len:
      logger.warning(
        "Token indices sequence length is longer than the specified maximum "
        " sequence length for this BERT model ({} > {}). Running this"
        " sequence through BERT will result in indexing errors".format(len(ids), self.max_len)
      )
    return ids

  def convert_ids_to_tokens(self, ids):
    """Converts a sequence of ids in wordpiece tokens using the vocab."""
    tokens = []
    for i in ids:
      tokens.append(self.ids_to_tokens[i])
    return tokens

  @classmethod
  def from_pretrained(cls, pretrained_model_name_or_path, cache_dir=None, *inputs, **kwargs):
    """
    Instantiate a PreTrainedBertModel from a pre-trained model file.
    Download and cache the pre-trained model file if needed.
    """
    if pretrained_model_name_or_path in PRETRAINED_VOCAB_ARCHIVE_MAP:
      vocab_file = PRETRAINED_VOCAB_ARCHIVE_MAP[pretrained_model_name_or_path]
    else:
      vocab_file = pretrained_model_name_or_path
    if os.path.isdir(vocab_file):
      vocab_file = os.path.join(vocab_file, VOCAB_NAME)
    # redirect to the cache, if necessary
    try:
      resolved_vocab_file = cached_path(vocab_file, cache_dir=cache_dir)
    except EnvironmentError:
      logger.error(
        "Model name '{}' was not found in model name list ({}). "
        "We assumed '{}' was a path or url but couldn't find any file "
        "associated to this path or url.".format(
          pretrained_model_name_or_path,
          ', '.join(PRETRAINED_VOCAB_ARCHIVE_MAP.keys()),
          vocab_file))
      return None
    if resolved_vocab_file == vocab_file:
      logger.info("loading vocabulary file {}".format(vocab_file))
    else:
      logger.info("loading vocabulary file {} from cache at {}".format(
        vocab_file, resolved_vocab_file))
    if pretrained_model_name_or_path in PRETRAINED_VOCAB_POSITIONAL_EMBEDDINGS_SIZE_MAP:
      # if we're using a pretrained model, ensure the tokenizer wont index sequences longer
      # than the number of positional embeddings
      max_len = PRETRAINED_VOCAB_POSITIONAL_EMBEDDINGS_SIZE_MAP[pretrained_model_name_or_path]
      kwargs['max_len'] = min(kwargs.get('max_len', int(1e12)), max_len)
    # Instantiate tokenizer.
    tokenizer = cls(resolved_vocab_file, *inputs, **kwargs)
    return tokenizer


class BasicTokenizer(object):
  """Runs basic tokenization (punctuation splitting, lower casing, etc.)."""

  def __init__(self,
               do_lower_case=True,
               never_split=("[UNK]", "[SEP]", "[PAD]", "[CLS]", "[MASK]"),
               lang="", pretokenized=False):
    """Constructs a BasicTokenizer.

    Args:
      do_lower_case: Whether to lower case the input.
    """
    self.do_lower_case = do_lower_case
    self.never_split = never_split
    self.lang = lang
    self.pretokenized = pretokenized

  def tokenize(self, text):
    """Tokenizes a piece of text."""
    is_head = []
    if not self.pretokenized:
      text = self._clean_text(text)

    # This was added on November 1st, 2018 for the multilingual and Chinese
    # models. This is also applied to the English models now, but it doesn't
    # matter since the English models were not trained on any Chinese data
    # and generally don't have any Chinese data in them (there are Chinese
    # characters in the vocabulary because Wikipedia does have some Chinese
    # words in the English Wikipedia.).
    if 'zh' in self.lang:
      text = self._tokenize_chinese_chars(text)
    orig_tokens = whitespace_tokenize(text)
    split_tokens = []
    for token in orig_tokens:

      if self.do_lower_case and token not in self.never_split:
        token = token.lower()
        token = self._run_strip_accents(token)
      if not self.pretokenized:
        splits = self._run_split_on_punc(token)
      else:
        splits = [token]
      for idx, s in enumerate(splits):
        if idx == 0:
          is_head.append(1)
        else:
          is_head.append(0)
      split_tokens.extend(splits)

    output_tokens = whitespace_tokenize(" ".join(split_tokens))
    return output_tokens, is_head

  def _run_strip_accents(self, text):
    """Strips accents from a piece of text."""
    text = unicodedata.normalize("NFD", text)
    output = []
    for char in text:
      cat = unicodedata.category(char)
      if cat == "Mn":
        continue
      output.append(char)
    return "".join(output)

  def _run_split_on_punc(self, text):
    """Splits punctuation on a piece of text."""
    if text in self.never_split:
      return [text]
    chars = list(text)
    i = 0
    start_new_word = True
    output = []
    while i < len(chars):
      char = chars[i]
      if _is_punctuation(char):
        output.append([char])
        start_new_word = True
      else:
        if start_new_word:
          output.append([])
        start_new_word = False
        output[-1].append(char)
      i += 1

    return ["".join(x) for x in output]

  def _tokenize_chinese_chars(self, text):
    """Adds whitespace around any CJK character."""
    output = []
    for char in text:
      cp = ord(char)
      if self._is_chinese_char(cp):
        output.append(" ")
        output.append(char)
        output.append(" ")
      else:
        output.append(char)
    return "".join(output)

  def _is_chinese_char(self, cp):
    """Checks whether CP is the codepoint of a CJK character."""
    # This defines a "chinese character" as anything in the CJK Unicode block:
    #   https://en.wikipedia.org/wiki/CJK_Unified_Ideographs_(Unicode_block)
    #
    # Note that the CJK Unicode block is NOT all Japanese and Korean characters,
    # despite its name. The modern Korean Hangul alphabet is a different block,
    # as is Japanese Hiragana and Katakana. Those alphabets are used to write
    # space-separated words, so they are not treated specially and handled
    # like the all of the other languages.
    if ((cp >= 0x4E00 and cp <= 0x9FFF) or  #
          (cp >= 0x3400 and cp <= 0x4DBF) or  #
          (cp >= 0x20000 and cp <= 0x2A6DF) or  #
          (cp >= 0x2A700 and cp <= 0x2B73F) or  #
          (cp >= 0x2B740 and cp <= 0x2B81F) or  #
          (cp >= 0x2B820 and cp <= 0x2CEAF) or
          (cp >= 0xF900 and cp <= 0xFAFF) or  #
          (cp >= 0x2F800 and cp <= 0x2FA1F)):  #
      return True

    return False

  def _clean_text(self, text):
    """Performs invalid character removal and whitespace cleanup on text."""
    output = []
    for char in text:
      cp = ord(char)
      if cp == 0 or cp == 0xfffd or _is_control(char):
        continue
      if _is_whitespace(char):
        output.append(" ")
      else:
        output.append(char)
    return "".join(output)


class WordpieceTokenizer(object):
  """Runs WordPiece tokenization."""

  def __init__(self, vocab, unk_token="[UNK]", max_input_chars_per_word=100):
    self.vocab = vocab
    self.unk_token = unk_token
    self.max_input_chars_per_word = max_input_chars_per_word

  def tokenize(self, text):
    """Tokenizes a piece of text into its word pieces.

    This uses a greedy longest-match-first algorithm to perform tokenization
    using the given vocabulary.

    For example:
      input = "unaffable"
      output = ["un", "##aff", "##able"]

    Args:
      text: A single token or whitespace separated tokens. This should have
        already been passed through `BasicTokenizer`.

    Returns:
      A list of wordpiece tokens.
    """

    output_tokens = []
    for token in whitespace_tokenize(text):
      chars = list(token)
      if len(chars) > self.max_input_chars_per_word:
        output_tokens.append(self.unk_token)
        continue

      is_bad = False
      start = 0
      sub_tokens = []
      while start < len(chars):
        end = len(chars)
        cur_substr = None
        while start < end:
          substr = "".join(chars[start:end])
          if start > 0:
            substr = "##" + substr
          if substr in self.vocab:
            cur_substr = substr
            break
          end -= 1
        if cur_substr is None:
          is_bad = True
          break
        sub_tokens.append(cur_substr)
        start = end

      if is_bad:
        output_tokens.append(self.unk_token)
      else:
        output_tokens.extend(sub_tokens)
    return output_tokens


def _is_whitespace(char):
  """Checks whether `chars` is a whitespace character."""
  # \t, \n, and \r are technically contorl characters but we treat them
  # as whitespace since they are generally considered as such.
  if char == " " or char == "\t" or char == "\n" or char == "\r":
    return True
  cat = unicodedata.category(char)
  if cat == "Zs":
    return True
  return False


def _is_control(char):
  """Checks whether `chars` is a control character."""
  # These are technically control characters but we count them as whitespace
  # characters.
  if char == "\t" or char == "\n" or char == "\r":
    return False
  cat = unicodedata.category(char)
  if cat.startswith("C"):
    return True
  return False


def _is_punctuation(char):
  """Checks whether `chars` is a punctuation character."""
  cp = ord(char)
  # We treat all non-letter/number ASCII as punctuation.
  # Characters such as "^", "$", and "`" are not in the Unicode
  # Punctuation class but we treat them as punctuation anyways, for
  # consistency.
  if ((cp >= 33 and cp <= 47) or (cp >= 58 and cp <= 64) or
        (cp >= 91 and cp <= 96) or (cp >= 123 and cp <= 126)):
    return True
  cat = unicodedata.category(char)
  if cat.startswith("P"):
    return True
  return False



import json
from types import SimpleNamespace

import relogic.utils.crash_on_ipy
from relogic.logickit.base.constants import SEQUENCE_CLASSIFICATION_TASK
from relogic.logickit.dataflow import TASK_TO_DATAFLOW_CLASS_MAP, SingletonDataFlow
from relogic.logickit.tokenizer.tokenization import BertTokenizer

config = SimpleNamespace(
  **{
    "buckets": [(0, 15), (15, 40), (40, 450)],
    "max_seq_length": 450,
    "label_mapping_path": "data/preprocessed_data/entity_type_classification.json"
  })

tokenizers = {
  "BPE": BertTokenizer.from_pretrained("bert-base-multilingual-cased", do_lower_case=False),
}

dataflow: SingletonDataFlow = TASK_TO_DATAFLOW_CLASS_MAP[SEQUENCE_CLASSIFICATION_TASK](
  task_name=SEQUENCE_CLASSIFICATION_TASK,
  config=config,
  tokenizers=tokenizers,
  label_mapping=json.load(open(config.label_mapping_path)))

examples = [{
  "tokens": ["I", "visited", "China", "yesterday", "."],
  "labels": ["LOC"]
},{
  "tokens": ["Barack", "Obama", "went", "to", "Paris", "."],
  "labels": ["PER", "LOC"]
}]

dataflow.update_with_jsons(examples)

for mb in dataflow.get_minibatches(minibatch_size=2):
  print(mb)

raise NotImplementedError("You can start to play with data")


import abc
from relogic.logickit.scorer.scorer import Scorer

class SentLevelScorer(Scorer, metaclass=abc.ABCMeta):
  def __init__(self):
    super(SentLevelScorer, self).__init__()
    self._total_loss = 0
    self._total_sents = 0
    self._examples = []
    self._preds = []
    self._attn_maps = []


  # def update(self, mbs, predictions, loss, extra_args):
  #   super(SentLevelScorer, self).update(mbs, predictions, loss, extra_args)
  #   n_sents = 0
  #   for example, preds in zip(mbs.examples, predictions):
  #     self._examples.append(example)
  #     self._preds.append(preds)
  #     # self._attn_maps.append(attention_map)
  #     # do not dumping heavy results on memory
  #     n_sents += 1
  #   self._total_loss += loss * n_sents
  #   self._total_sents += n_sents

  def get_loss(self):
    return self._total_loss / max(1, self._total_sents)


import torch.nn as nn
import torch
from relogic.logickit.modules.initializers import block_orthogonal
from relogic.logickit.modules.rnn import dynamic_rnn

class LSTM(nn.Module):
  """
  Args:
    input_size: The number of expected features in the input `x`
    hidden_size: The number of features in the hidden state `h`
    num_layers: Number of recurrent layers. E.g., setting ``num_layers=2``
        would mean stacking two LSTMs together to form a `stacked LSTM`,
        with the second LSTM taking in outputs of the first LSTM and
        computing the final results. Default: 1
    bias: If ``False``, then the layer does not use bias weights `b_ih` and `b_hh`.
        Default: ``True``
    batch_first: If ``True``, then the input and output tensors are provided
        as (batch, seq, feature). Default: ``False``
    dropout: If non-zero, introduces a `Dropout` layer on the outputs of each
        LSTM layer except the last layer, with dropout probability equal to
        :attr:`dropout`. Default: 0
    bidirectional: If ``True``, becomes a bidirectional LSTM. Default: ``False``

  Attributes:
    weight_ih_l[k] : the learnable input-hidden weights of the :math:`\text{k}^{th}` layer
        `(W_ii|W_if|W_ig|W_io)`, of shape `(4*hidden_size, input_size)` for `k = 0`.
        Otherwise, the shape is `(4*hidden_size, num_directions * hidden_size)`
    weight_hh_l[k] : the learnable hidden-hidden weights of the :math:`\text{k}^{th}` layer
        `(W_hi|W_hf|W_hg|W_ho)`, of shape `(4*hidden_size, hidden_size)`
    bias_ih_l[k] : the learnable input-hidden bias of the :math:`\text{k}^{th}` layer
        `(b_ii|b_if|b_ig|b_io)`, of shape `(4*hidden_size)`
    bias_hh_l[k] : the learnable hidden-hidden bias of the :math:`\text{k}^{th}` layer
        `(b_hi|b_hf|b_hg|b_ho)`, of shape `(4*hidden_size)`

  """
  def __init__(self,
               input_size,
               hidden_size,
               num_layers=1,
               bias=True,
               batch_first=False,
               dropout=0,
               bidirectional=False):
    super(LSTM, self).__init__()
    self.lstm = nn.LSTM(input_size, hidden_size, num_layers,
      bias, batch_first, dropout, bidirectional)
    for layer in self.lstm._all_weights:
      for param in layer:
        if 'weight_ih' in param:
          block_orthogonal(
            tensor=self.lstm.__getattr__(param),
            split_sizes=[hidden_size, input_size])
        if 'weight_hh' in param:
          block_orthogonal(
            tensor=self.lstm.__getattr__(param),
            split_sizes=[hidden_size, hidden_size])

  def forward(self, inputs, lengths=None):
    if lengths is None:
      return self.lstm(inputs)
    else:
      return dynamic_rnn(self.lstm, inputs, lengths)

class HighwayLSTM(nn.Module):
  def __init__(self, num_layers, input_size, hidden_size, layer_dropout):
    super(HighwayLSTM, self).__init__()
    self.num_layers = num_layers
    self.layers = nn.ModuleList([LSTM(
      input_size=input_size if layer == 0 else hidden_size * 2,  # for indicator embedding
      hidden_size=hidden_size,
      bidirectional=True,
      batch_first=True,
      num_layers=1
    ) for layer in range(num_layers)])
    self.dropout = nn.Dropout(p=layer_dropout)
    self.linear_layers = nn.ModuleList([torch.nn.Linear(hidden_size * 2, hidden_size * 2)
                                        for _ in range(num_layers - 1)])


  def forward(self, inputs, lengths):
    current_inputs = inputs
    for idx, layer in enumerate(self.layers):
      output, _ = layer(current_inputs, lengths)
      output = self.dropout(output)
      if idx > 0:
        gate = self.linear_layers[idx-1](output)
        gate = torch.sigmoid(gate)
        output = gate * output + (1 - gate) * current_inputs
      current_inputs = output
    return output


import argparse
import collections
import random


def main(input_file, output_file, num_neg_per_qurey):
  example_collections = collections.defaultdict(list)
  with open(input_file) as fin:
    for line in fin:
      items = line.strip().split('\t')
      pair_id = items[0].split('-')
      pos_id = pair_id[0]
      example_collections[pos_id].append(line)
  with open(output_file, "w") as fout:
    for key in example_collections:
      samples = random.sample(example_collections[key], min(num_neg_per_qurey, len(example_collections[key])))
      for sample in samples:
        fout.write(sample)

if __name__ == "__main__":
  parser = argparse.ArgumentParser() 
  parser.add_argument("--input_file", type=str)
  parser.add_argument("--num_neg_per_query", type=int, default=15)
  parser.add_argument("--output_file", type=str)
  args = parser.parse_args()
  main(args.input_file, args.output_file, args.num_neg_per_query)


class PrefixMap(object):
  def __init__(self, mapping):
    self.mapping = mapping

  def __getitem__(self, item):
    selected_key, data = None, None
    for key in self.mapping:
      if item == key:
        data = self.mapping[key]
    if data is None:
      for key in self.mapping:
        if item.startswith(key):
          if data is None:
            data = self.mapping[key]
          else:
            raise ValueError("{} is a common prefix of {} and {}".format(item, selected_key, key))
    if data is None:
      raise ValueError("{} is not in the map".format(item))
    return data



from dataclasses import dataclass, field
from typing import List

from relogic.structures.structure import Structure
from relogic.structures.sentence import Sentence
from nltk.tokenize import sent_tokenize

@dataclass
class Paragraph(Structure):
  idx: int = None
  text: str= None
  sentences: List[Sentence] = field(default_factory=list)

  tokenizer: str = "space"

  def __post_init__(self):
    for sent in sent_tokenize(self.text):
      self.add_sentence(Sentence(text=sent, tokenizer=self.tokenizer))

  def add_sentence(self, sentence: Sentence):
    self.sentences.append(sentence)


import torch.nn as nn
import torch
from relogic.logickit.modules.rnn import dynamic_rnn
import torch.nn.functional as F
from relogic.logickit.modules.span_gcn import select_span

class PredictionModule(nn.Module):
  def __init__(self, config, task_name, n_classes):
    super(PredictionModule, self).__init__()
    self.config = config
    self.task_name = task_name
    self.n_classes = n_classes
    if config.use_bilstm:
      self.bilstm = nn.LSTM(
        input_size=config.hidden_size, # for indicator embedding
        hidden_size=config.hidden_size,
        bidirectional=False,
        batch_first=True,
        num_layers=1)
      self.projection = nn.Linear(config.hidden_size, config.hidden_size)
      self.activation = nn.SELU()
      self.attention_W = nn.Linear(2 * config.hidden_size + 20, config.hidden_size, bias=False)
      self.attention_V = nn.Linear(config.hidden_size, 1, bias=False)
    if self.config.rel_extraction_module_type == "entity_span":
      hidden_size = 2 * config.hidden_size
    elif self.config.rel_extraction_module_type == "sent_repr":
      hidden_size = config.hidden_size
    elif self.config.rel_extraction_module_type == "hybrid":
      hidden_size = 3 * config.hidden_size
    else:
      raise ValueError("rel_extraction_module_type is not in [entity_span, sent_repr, hybrid], {}".format(
        self.config.rel_extraction_module_type))
    self.to_logits = nn.Linear(hidden_size, self.n_classes)
    if self.config.rel_extraction_module_type == "hybrid":
      self.padding = nn.Parameter(torch.zeros(config.hidden_size), requires_grad=False)

  def attention_module(self, query, candidate, mask):
    # query = (batch, dim), candidate = (batch, length, dim)
    query_duplicate = query.unsqueeze(1).expand(-1, candidate.size(1), -1)
    mask = mask.to(dtype=next(self.parameters()).dtype)
    u = self.attention_V(
      F.tanh(
        self.attention_W(torch.cat([query_duplicate, candidate], dim=-1))
      )).squeeze(-1) + (1.0-mask) * (-10000.0)
    # u = (batch, length)
    a = F.softmax(u, dim=-1)
    return a

  def forward(self, input, subj_indicator_embed, obj_indicator_embed, input_lengths, input_mask, extra_args, **kwargs):
    if self.config.use_bilstm:
      # input_concat = torch.cat([input, subj_indicator_embed, obj_indicator_embed], dim=-1)
      output, (ht, ct) = dynamic_rnn(self.bilstm, input, input_lengths)
      # ht (2, batch, dim)
      ht = ht.transpose(0, 1).contiguous().view(input.size(0), -1)
      # (batch_size, dim)
      prob = self.attention_module(
        query=ht,
        candidate=torch.cat([output, subj_indicator_embed, obj_indicator_embed], dim=-1),
        mask=input_mask)
      projected = (output * prob.unsqueeze(-1)).sum(1)
      # output (batch, sequence, dim) -> projected (batch, dim)
      projected = self.projection(projected)
      projected = self.activation(projected)
    else:
      if self.config.rel_extraction_module_type == "entity_span":
        if "start_of_subject" in extra_args and "start_of_object" in extra_args:
          start_of_subject = extra_args["start_of_subject"]
          start_of_object = extra_args["start_of_object"]
          start_of_subject_repr = input[torch.arange(input.size(0)), start_of_subject]
          start_of_object_repr = input[torch.arange(input.size(0)), start_of_object]
          projected = torch.cat([start_of_subject_repr, start_of_object_repr], dim=1)
          # logits = self.to_logits(torch.cat([start_of_subject_repr, start_of_object_repr], dim=1))
        else:
          raise ValueError("Start of Subject is not in extra args")
      elif self.config.rel_extraction_module_type == "sent_repr":
        projected = input[:, 0]
      elif self.config.rel_extraction_module_type == "hybrid":
        if ("start_of_subject" in extra_args and
              "start_of_object" in extra_args and
              "end_of_subject" in extra_args and
              "end_of_object" in extra_args):
          cls_hidden = input[:, 0]
          start_of_subject = extra_args["start_of_subject"]
          start_of_object = extra_args["start_of_object"]
          end_of_subject = extra_args["end_of_subject"]
          end_of_object = extra_args["end_of_object"]
          subject_repr = select_span(input, start_index=start_of_subject, end_index=end_of_subject, padding=self.padding)
          object_repr = select_span(input, start_index=start_of_object, end_index=end_of_object, padding=self.padding)
          subject_repr = self.agg(subject_repr, end_of_subject-start_of_subject)
          object_repr = self.agg(object_repr, end_of_object-start_of_object)
          projected = torch.cat([cls_hidden, subject_repr, object_repr], dim=-1)
        else:
          raise ValueError("Start of Subject is not in extra args")
      else:
        raise ValueError("rel_extraction_module_type is not in [entity_span, sent_repr, hybrid], {}".format(
          self.config.rel_extraction_module_type))
    logits = self.to_logits(projected)
    return logits

  def agg(self, embed, lengths):
    return torch.sum(embed, -2) / lengths.unsqueeze(1).float()



class RelExtractionModule(nn.Module):
  def __init__(self, config, task_name, n_classes):
    super(RelExtractionModule, self).__init__()
    self.config = config
    self.indicator_embedding = nn.Embedding(self.config.max_seq_length * 2 + 1, 10)

    self.primary = PredictionModule(config, task_name, n_classes)

  def forward(self,
              input,
              view_type="primary",
              final_multi_head_repr=None,
              input_mask=None,
              segment_ids=None,
              extra_args=None):
    if view_type == "primary":
      input_lengths = (segment_ids == 0).sum(-1)
      subj_ind = extra_args['subj_indicator'] + self.config.max_seq_length # torch.tensor(self.config.max_seq_length, dtype=torch.long).to(input.device)
      obj_ind = extra_args['obj_indicator'] + self.config.max_seq_length # torch.tensor(self.config.max_seq_length, dtype=torch.long).to(input.device)
      subj_indicator_embed = self.indicator_embedding(subj_ind.unsqueeze(1))
      obj_indicator_embed = self.indicator_embedding(obj_ind.unsqueeze(1))
      primary_logits = self.primary(input, subj_indicator_embed, obj_indicator_embed, input_lengths, input_mask, extra_args)
      return primary_logits
    else:
      raise NotImplementedError("Partial View is not implemented")


import torch
import torch.nn as nn
import torch.nn.functional as F


class SimpleCNN(nn.Module):
  def __init__(self, num_of_conv, in_channels, out_channels, kernel_size, in_features, out_features=None, stride=1,
               dilation=1, groups=1, bias=True, active_func=F.relu, pooling=F.max_pool1d,
               dropout=0.5, padding_strategy="default", padding_list=None, fc_layer=True,
               include_map=False, k_max_pooling=False, k=1):
    """
    :param num_of_conv: Follow kim cnn idea
    :param kernel_size: if is int type, then make it into list, length equals to num_of_conv
                 if list type, then check the length of it, should has length of num_of_conv
    :param out_features: feature size
    """
    super(SimpleCNN, self).__init__()
    if type(kernel_size) == int:
      kernel_size = [kernel_size]
    if len(kernel_size) != num_of_conv:
      print("Number of kernel_size should be same num_of_conv")
      exit(1)
    if padding_list == None:
      if padding_strategy == "default":
        padding_list = [(k_size - 1, 0) for k_size in kernel_size]
    self.include_map = include_map

    self.conv = nn.ModuleList([nn.Conv2d(in_channels=in_channels,
                                         out_channels=out_channels,
                                         kernel_size=(k_size, in_features),
                                         stride=stride,
                                         padding=padding,
                                         dilation=dilation,
                                         groups=groups,
                                         bias=bias)
                               for k_size, padding in zip(kernel_size, padding_list)])
    self.pooling = pooling
    self.k_max_pooling = k_max_pooling
    self.k = k if k_max_pooling else 1
    self.active_func = active_func
    self.fc_layer = fc_layer
    if fc_layer:
      self.dropout = nn.Dropout(dropout)
      self.fc = nn.Linear(num_of_conv * out_channels * self.k, out_features)

  def kmax_pooling(self, x, dim, k):
    index = x.topk(k, dim=dim)[1].sort(dim=dim)[0]
    return x.gather(dim, index)

  def forward(self, input):
    batch_size = input.size(0)
    if len(input.size()) == 3:
      input = input.unsqueeze(1)
    # input = (batch, in_channels, sent_len, word_dim)
    x_map = [self.active_func(conv(input)).squeeze(3) for conv in self.conv]
    # (batch, channel_output, ~=sent_len) * Ks
    if self.k_max_pooling:
      x = [self.kmax_pooling(i, 2, self.k).view(batch_size, -1) for i in x_map]
    else:
      x = [self.pooling(i, i.size(2)).squeeze(2) for i in x_map]  # max-over-time pooling
    x = torch.cat(x, 1)  # (batch, out_channels * Ks)
    if self.fc_layer:
      x = self.dropout(x)
      x = self.fc(x)
    if self.include_map == False:
      return x
    else:
      return x, x_map

class SamCNN(nn.Module):
  def __init__(self, config, task_name, n_classes):
    super().__init__()

    self.sm_cnn = SimpleCNN(
      num_of_conv=1,
      in_channels=1,
      out_channels=config.output_channel,
      kernel_size=[config.kernel_size],
      in_features=config.word_embed_dim,
      out_features=config.hidden_size,
      active_func=nn.ReLU(),
      dropout=config.dropout,
      fc_layer=True
    )

    self.projection = nn.Linear(config.hidden_size * 3,
                config.projection_size)
    self.batch_norm = nn.BatchNorm1d(config.projection_size)
    self.to_logits = nn.Sequential(
      nn.Tanh(),
      nn.Dropout(config.dropout),
      nn.Linear(config.projection_size, n_classes))

    self.context_cnn = SimpleCNN(
      num_of_conv=1,
      in_channels=1,
      out_channels=config.output_channel,
      kernel_size=[config.kernel_size],
      in_features=config.word_embed_dim,
      out_features=config.hidden_size,
      active_func=nn.ReLU(),
      dropout=config.dropout,
      fc_layer=True
    )
    self.cos = nn.CosineSimilarity(dim=1, eps=1e-6)

  def attn_context_ave(self, question_embed, answer_embed, batch_size):
    question_len = question_embed.size(1)
    answer_len = answer_embed.size(1)
    dimension = question_embed.size(2)
    question = torch.cat([question_embed.unsqueeze(2)] * answer_len, dim=2).view(-1, dimension)
    answer = torch.cat([answer_embed.unsqueeze(1)] * question_len, dim=1).view(-1, dimension)
    # (batch, question_len, answer_len, dim)
    attn_prob = self.cos(answer, question).unsqueeze(1)
    attn_answer = (answer * attn_prob).view(batch_size * question_len, answer_len, dimension)
    feature = self.context_cnn(attn_answer).view(batch_size, question_len, -1)
    feature = torch.sum(feature, dim=1) / question_len
    return feature

  def forward(self, *inputs, **kwargs):
    a_features = kwargs.pop("a_features")
    b_features = kwargs.pop("b_features")

    a_cnn_features = self.sm_cnn(a_features)
    b_cnn_features = self.sm_cnn(b_features)
    attention_features = self.attn_context_ave(a_features, b_features, a_features.size(0))

    feat_comb = torch.cat([a_cnn_features, b_cnn_features, attention_features], dim=1)
    feat_comb = self.projection(feat_comb)
    if feat_comb.size(0) > 1:
      feat_comb = self.batch_norm(feat_comb)
    logits = self.to_logits(feat_comb)

    return logits






import os
import sys
from relogic.utils.file_utils import cached_path, RELOGIC_CACHE



PACKAGE_PATH = {
  "Anserini": "https://git.uwaterloo.ca/p8shi/data-server/raw/master/anserini-0.6.0-SNAPSHOT-fatjar.jar"
}

anserini_cache_path = cached_path(PACKAGE_PATH['Anserini'], cache_dir=RELOGIC_CACHE)


if sys.platform == 'win32':
  separator = ';'
else:
  separator = ':'

jar = os.path.join(separator + anserini_cache_path)

if 'CLASSPATH' not in os.environ:
  os.environ['CLASSPATH'] = jar
else:
  os.environ['CLASSPATH'] += jar


import tensorflow as tf
from relogic.tpukit.tokenizer import tokenization
import collections


def _decode_record(record, name_to_features):
  """Decodes a record to a TensorFlow example."""
  example = tf.parse_single_example(record, name_to_features)

  # tf.Example only supports tf.int64, but the TPU only supports tf.int32.
  # So cast all int64 to int32.
  for name in list(example.keys()):
    t = example[name]
    if t.dtype == tf.int64:
      t = tf.to_int32(t)
    example[name] = t

  return example


def input_fn_builder(input_files,
                     max_seq_length,
                     max_predictions_per_seq,
                     is_training,
                     num_cpu_threads=4):
  """Creates an `input_fn` closure to be passed to TPUEstimator."""

  def input_fn(params):
    """The actual input function."""
    batch_size = params["batch_size"]

    name_to_features = {
        "input_ids":
            tf.FixedLenFeature([max_seq_length], tf.int64),
        "input_mask":
            tf.FixedLenFeature([max_seq_length], tf.int64),
        "segment_ids":
            tf.FixedLenFeature([max_seq_length], tf.int64),
        "masked_lm_positions":
            tf.FixedLenFeature([max_predictions_per_seq], tf.int64),
        "masked_lm_ids":
            tf.FixedLenFeature([max_predictions_per_seq], tf.int64),
        "masked_lm_weights":
            tf.FixedLenFeature([max_predictions_per_seq], tf.float32),
        "next_sentence_labels":
            tf.FixedLenFeature([1], tf.int64),
    }

    # For training, we want a lot of parallel reading and shuffling.
    # For eval, we want no shuffling and parallel reading doesn't matter.
    if is_training:
      d = tf.data.Dataset.from_tensor_slices(tf.constant(input_files))
      d = d.repeat()
      d = d.shuffle(buffer_size=len(input_files))

      # `cycle_length` is the number of parallel files that get read.
      cycle_length = min(num_cpu_threads, len(input_files))

      # `sloppy` mode means that the interleaving is not exact. This adds
      # even more randomness to the training pipeline.
      d = d.apply(
          tf.contrib.data.parallel_interleave(
              tf.data.TFRecordDataset,
              sloppy=is_training,
              cycle_length=cycle_length))
      d = d.shuffle(buffer_size=100)
    else:
      d = tf.data.TFRecordDataset(input_files)
      # Since we evaluate for a fixed number of steps we don't want to encounter
      # out-of-range exceptions.
      d = d.repeat()

    # We must `drop_remainder` on training because the TPU requires fixed
    # size dimensions. For eval, we assume we are evaluating on the CPU or GPU
    # and we *don't* want to drop the remainder, otherwise we wont cover
    # every sample.
    d = d.apply(
        tf.contrib.data.map_and_batch(
            lambda record: _decode_record(record, name_to_features),
            batch_size=batch_size,
            num_parallel_batches=num_cpu_threads,
            drop_remainder=True))
    return d

  return input_fn


def file_based_input_fn_builder(input_file, seq_length, is_training,
                                drop_remainder):
  """Creates an `input_fn` closure to be passed to TPUEstimator."""

  name_to_features = {
      "input_ids": tf.FixedLenFeature([seq_length], tf.int64),
      "input_mask": tf.FixedLenFeature([seq_length], tf.int64),
      "segment_ids": tf.FixedLenFeature([seq_length], tf.int64),
      "label_ids": tf.FixedLenFeature([], tf.int64),
      "is_real_example": tf.FixedLenFeature([], tf.int64),
  }

  def _decode_record(record, name_to_features):
    """Decodes a record to a TensorFlow example."""
    example = tf.parse_single_example(record, name_to_features)

    # tf.Example only supports tf.int64, but the TPU only supports tf.int32.
    # So cast all int64 to int32.
    for name in list(example.keys()):
      t = example[name]
      if t.dtype == tf.int64:
        t = tf.to_int32(t)
      example[name] = t

    return example

  def input_fn(params):
    """The actual input function."""
    batch_size = params["batch_size"]

    # For training, we want a lot of parallel reading and shuffling.
    # For eval, we want no shuffling and parallel reading doesn't matter.
    d = tf.data.TFRecordDataset(input_file)
    if is_training:
      d = d.repeat()
      d = d.shuffle(buffer_size=100)

    d = d.apply(
        tf.contrib.data.map_and_batch(
            lambda record: _decode_record(record, name_to_features),
            batch_size=batch_size,
            drop_remainder=drop_remainder))

    return d

  return input_fn


def convert_single_example(ex_index, example, label_list, max_seq_length,
                           tokenizer):
  """Converts a single `InputExample` into a single `InputFeatures`."""

  if isinstance(example, PaddingInputExample):
    return InputFeatures(
        input_ids=[0] * max_seq_length,
        input_mask=[0] * max_seq_length,
        segment_ids=[0] * max_seq_length,
        label_id=0,
        is_real_example=False)

  label_map = {}
  for (i, label) in enumerate(label_list):
    label_map[label] = i

  tokens_a = tokenizer.tokenize(example.text_a)
  tokens_b = None
  if example.text_b:
    tokens_b = tokenizer.tokenize(example.text_b)

  if tokens_b:
    # Modifies `tokens_a` and `tokens_b` in place so that the total
    # length is less than the specified length.
    # Account for [CLS], [SEP], [SEP] with "- 3"
    _truncate_seq_pair(tokens_a, tokens_b, max_seq_length - 3)
  else:
    # Account for [CLS] and [SEP] with "- 2"
    if len(tokens_a) > max_seq_length - 2:
      tokens_a = tokens_a[0:(max_seq_length - 2)]

  # The convention in BERT is:
  # (a) For sequence pairs:
  #  tokens:   [CLS] is this jack ##son ##ville ? [SEP] no it is not . [SEP]
  #  type_ids: 0     0  0    0    0     0       0 0     1  1  1  1   1 1
  # (b) For single sequences:
  #  tokens:   [CLS] the dog is hairy . [SEP]
  #  type_ids: 0     0   0   0  0     0 0
  #
  # Where "type_ids" are used to indicate whether this is the first
  # sequence or the second sequence. The embedding vectors for `type=0` and
  # `type=1` were learned during pre-training and are added to the wordpiece
  # embedding vector (and position vector). This is not *strictly* necessary
  # since the [SEP] token unambiguously separates the sequences, but it makes
  # it easier for the model to learn the concept of sequences.
  #
  # For classification tasks, the first vector (corresponding to [CLS]) is
  # used as the "sentence vector". Note that this only makes sense because
  # the entire model is fine-tuned.
  tokens = []
  segment_ids = []
  tokens.append("[CLS]")
  segment_ids.append(0)
  for token in tokens_a:
    tokens.append(token)
    segment_ids.append(0)
  tokens.append("[SEP]")
  segment_ids.append(0)

  if tokens_b:
    for token in tokens_b:
      tokens.append(token)
      segment_ids.append(1)
    tokens.append("[SEP]")
    segment_ids.append(1)

  input_ids = tokenizer.convert_tokens_to_ids(tokens)

  # The mask has 1 for real tokens and 0 for padding tokens. Only real
  # tokens are attended to.
  input_mask = [1] * len(input_ids)

  # Zero-pad up to the sequence length.
  while len(input_ids) < max_seq_length:
    input_ids.append(0)
    input_mask.append(0)
    segment_ids.append(0)

  assert len(input_ids) == max_seq_length
  assert len(input_mask) == max_seq_length
  assert len(segment_ids) == max_seq_length

  label_id = label_map[example.label]
  if ex_index < 5:
    tf.logging.info("*** Example ***")
    tf.logging.info("guid: %s" % (example.guid))
    tf.logging.info("tokens: %s" % " ".join(
        [tokenization.printable_text(x) for x in tokens]))
    tf.logging.info("input_ids: %s" % " ".join([str(x) for x in input_ids]))
    tf.logging.info("input_mask: %s" % " ".join([str(x) for x in input_mask]))
    tf.logging.info("segment_ids: %s" % " ".join([str(x) for x in segment_ids]))
    tf.logging.info("label: %s (id = %d)" % (example.label, label_id))

  feature = InputFeatures(
      input_ids=input_ids,
      input_mask=input_mask,
      segment_ids=segment_ids,
      label_id=label_id,
      is_real_example=True)
  return feature


def file_based_convert_examples_to_features(
    examples, label_list, max_seq_length, tokenizer, output_file):
  """Convert a set of `InputExample`s to a TFRecord file."""

  writer = tf.python_io.TFRecordWriter(output_file)

  for (ex_index, example) in enumerate(examples):
    if ex_index % 10000 == 0:
      tf.logging.info("Writing example %d of %d" % (ex_index, len(examples)))

    feature = convert_single_example(ex_index, example, label_list,
                                     max_seq_length, tokenizer)

    def create_int_feature(values):
      f = tf.train.Feature(int64_list=tf.train.Int64List(value=list(values)))
      return f

    features = collections.OrderedDict()
    features["input_ids"] = create_int_feature(feature.input_ids)
    features["input_mask"] = create_int_feature(feature.input_mask)
    features["segment_ids"] = create_int_feature(feature.segment_ids)
    features["label_ids"] = create_int_feature([feature.label_id])
    features["is_real_example"] = create_int_feature(
        [int(feature.is_real_example)])

    tf_example = tf.train.Example(features=tf.train.Features(feature=features))
    writer.write(tf_example.SerializeToString())
  writer.close()

class InputExample(object):
  """A single training/test example for simple sequence classification."""

  def __init__(self, guid, text_a, text_b=None, label=None):
    """Constructs a InputExample.
    Args:
      guid: Unique id for the example.
      text_a: string. The untokenized text of the first sequence. For single
        sequence tasks, only this sequence must be specified.
      text_b: (Optional) string. The untokenized text of the second sequence.
        Only must be specified for sequence pair tasks.
      label: (Optional) string. The label of the example. This should be
        specified for train and dev examples, but not for test examples.
    """
    self.guid = guid
    self.text_a = text_a
    self.text_b = text_b
    self.label = label


class PaddingInputExample(object):
  """Fake example so the num input examples is a multiple of the batch size.
  When running eval/predict on the TPU, we need to pad the number of examples
  to be a multiple of the batch size, because the TPU requires a fixed batch
  size. The alternative is to drop the last batch, which is bad because it means
  the entire output data won't be generated.
  We use this class instead of `None` because treating `None` as padding
  battches could cause silent errors.
  """


class InputFeatures(object):
  """A single set of features of data."""

  def __init__(self,
               input_ids,
               input_mask,
               segment_ids,
               label_id,
               is_real_example=True):
    self.input_ids = input_ids
    self.input_mask = input_mask
    self.segment_ids = segment_ids
    self.label_id = label_id
    self.is_real_example = is_real_example


def _truncate_seq_pair(tokens_a, tokens_b, max_length):
  """Truncates a sequence pair in place to the maximum length."""

  # This is a simple heuristic which will always truncate the longer sequence
  # one token at a time. This makes more sense than truncating an equal percent
  # of tokens from each, since if one sequence is very short then each token
  # that's truncated likely contains more information than a longer sequence.
  while True:
    total_length = len(tokens_a) + len(tokens_b)
    if total_length <= max_length:
      break
    if len(tokens_a) > len(tokens_b):
      tokens_a.pop()
    else:
      tokens_b.pop()





import abc
from relogic.logickit.scorer.scorer import Scorer


class DistanceScorer(Scorer):
  def __init__(self):
    super(DistanceScorer, self).__init__()
    self.loss = 0
    self.batch_count = 0

  def update(self, mbs, predictions, loss, extra_args):
    super(DistanceScorer, self).update(mbs, predictions, loss, extra_args)
    self.loss += loss
    self.batch_count += 1

  def get_loss(self):
    return self.loss / self.batch_count

  def _get_results(self):
    return [("distance", - self.loss / self.batch_count)]



from typing import List, Tuple

import torch

from relogic.logickit.dataflow import DataFlow, Example, Feature, MiniBatch
from relogic.logickit.utils import create_tensor
from relogic.logickit.tokenizer import CustomizedBertTokenizer, RobertaXLMTokenizer


class MixSentExample(Example):
  """MixSent Example contains the attributes and functionality for
  MixSent Semi-supervised Learning.

  """
  def __init__(
        self, text_a: str, text_b: str, text_c: str,
        span_a, span_b, span_c_a, span_c_b):
    super().__init__()
    self.text_a = text_a
    self.a_raw_tokens = text_a.split()
    self.text_b = text_b
    self.b_raw_tokens = text_b.split()
    self.text_c = text_c
    self.c_raw_tokens = text_c.split()
    self.span_a = span_a
    self.span_b = span_b
    self.span_c_a = span_c_a
    self.span_c_b = span_c_b

    self.padding_id = 0

  def process(self, tokenizers, *inputs, **kwargs):
    for tokenizer in tokenizers.values():
      if isinstance(tokenizer, RobertaXLMTokenizer):
        (self.a_tokens, self.a_is_head,
         self.a_input_ids) = tokenizer.tokenize_and_add_placeholder_and_convert_to_ids(self.text_a, self.a_raw_tokens)
        (self.b_tokens, self.b_is_head,
         self.b_input_ids) = tokenizer.tokenize_and_add_placeholder_and_convert_to_ids(self.text_b, self.b_raw_tokens)
        (self.c_tokens, self.c_is_head,
         self.c_input_ids) = tokenizer.tokenize_and_add_placeholder_and_convert_to_ids(self.text_c, self.c_raw_tokens)

        self.a_segment_ids = [0] * len(self.a_tokens)
        self.b_segment_ids = [0] * len(self.b_tokens)
        self.c_segment_ids = [0] * len(self.c_tokens)

        self.a_head_index = [idx for idx, value in enumerate(self.a_is_head) if value == 1]
        self.b_head_index = [idx for idx, value in enumerate(self.b_is_head) if value == 1]
        self.c_head_index = [idx for idx, value in enumerate(self.c_is_head) if value == 1]

        self.a_input_mask = [1] * len(self.a_input_ids)
        self.b_input_mask = [1] * len(self.b_input_ids)
        self.c_input_mask = [1] * len(self.c_input_ids)

        self.span_a_selected_index = [self.a_head_index[index] for index in range(self.span_a[0], self.span_a[1])]
        self.span_b_selected_index = [self.b_head_index[index] for index in range(self.span_b[0], self.span_b[1])]
        self.span_c_a_selected_index = [self.c_head_index[index] for index in range(self.span_c_a[0], self.span_c_a[1])]
        self.span_c_b_selected_index = [self.c_head_index[index] for index in range(self.span_c_b[0], self.span_c_b[1])]

        self.padding_id = 1

  @classmethod
  def from_structure(cls, structure):
    pass

  @classmethod
  def from_json(cls, example):
    if isinstance(example["text_a"], list):
      # TODO depends on the language!
      text_a = " ".join(example["text_a"])
      text_b = " ".join(example["text_b"])
      text_c = " ".join(example["text_c"])
    elif isinstance(example["text_a"], str):
      text_a = example["text_a"]
      text_b = example["text_b"]
      text_c = example["text_c"]
    else:
      raise NotImplementedError("Unsupported data type {}".format(type(example["text_a"])))

    return cls(text_a=text_a,
               text_b=text_b,
               text_c=text_c,
               span_a=example["span_a"],
               span_b=example["span_b"],
               span_c_a=example["span_c_a"],
               span_c_b=example["span_c_b"])

  @property
  def len(self):
    return len(self.a_input_ids) + len(self.b_input_ids) + len(self.c_input_ids)

  @property
  def len_a(self):
    return len(self.a_input_ids)

  @property
  def len_b(self):
    return len(self.b_input_ids)

  @property
  def len_c(self):
    return len(self.c_input_ids)

class MixSentFeature(Feature):
  """

  """
  def __init__(self, *inputs, **kwargs):
    super().__init__()
    self.a_input_ids = kwargs.pop("a_input_ids")
    self.b_input_ids = kwargs.pop("b_input_ids")
    self.c_input_ids = kwargs.pop("c_input_ids")
    self.a_segment_ids = kwargs.pop("a_segment_ids")
    self.b_segment_ids = kwargs.pop("b_segment_ids")
    self.c_segment_ids = kwargs.pop("c_segment_ids")
    self.a_is_head = kwargs.pop("a_is_head")
    self.b_is_head = kwargs.pop("b_is_head")
    self.c_is_head = kwargs.pop("c_is_head")
    self.a_input_mask = kwargs.pop("a_input_mask")
    self.b_input_mask = kwargs.pop("b_input_mask")
    self.c_input_mask = kwargs.pop("c_input_mask")
    self.span_a_selected_index = kwargs.pop("span_a_selected_index")
    self.span_b_selected_index = kwargs.pop("span_b_selected_index")
    self.span_c_a_selected_index = kwargs.pop("span_c_a_selected_index")
    self.span_c_b_selected_index = kwargs.pop("span_c_b_selected_index")

class MixSentMiniBatch(MiniBatch):

  def __init__(self, *inputs, **kwargs):
    super().__init__(*inputs, **kwargs)

  def generate_input(self, device, use_label):
    inputs = {}
    inputs["task_name"] = self.task_name
    inputs["a_input_ids"] = create_tensor(self.input_features, "a_input_ids",
                                        torch.long, device)
    inputs["a_input_mask"] = create_tensor(self.input_features, "a_input_mask",
                                         torch.long, device)
    inputs["a_segment_ids"] = create_tensor(self.input_features, "a_segment_ids",
                                         torch.long, device)
    inputs["a_input_head"] = create_tensor(self.input_features, "a_is_head",
                                           torch.long, device)
    inputs["b_input_ids"] = create_tensor(self.input_features, "b_input_ids",
                                          torch.long, device)
    inputs["b_input_mask"] = create_tensor(self.input_features, "b_input_mask",
                                           torch.long, device)
    inputs["b_segment_ids"] = create_tensor(self.input_features, "b_segment_ids",
                                            torch.long, device)
    inputs["b_input_head"] = create_tensor(self.input_features, "b_is_head",
                                           torch.long, device)
    inputs["c_input_ids"] = create_tensor(self.input_features, "c_input_ids",
                                          torch.long, device)
    inputs["c_input_mask"] = create_tensor(self.input_features, "c_input_mask",
                                           torch.long, device)
    inputs["c_segment_ids"] = create_tensor(self.input_features, "c_segment_ids",
                                            torch.long, device)
    inputs["c_input_head"] = create_tensor(self.input_features, "c_is_head",
                                           torch.long, device)
    inputs["span_a_selected_index"] = create_tensor(self.input_features, "span_a_selected_index",
                                                    torch.long, device)
    inputs["span_b_selected_index"] = create_tensor(self.input_features, "span_b_selected_index",
                                                    torch.long, device)
    inputs["span_c_a_selected_index"] = create_tensor(self.input_features, "span_c_a_selected_index",
                                                    torch.long, device)
    inputs["span_c_b_selected_index"] = create_tensor(self.input_features, "span_c_b_selected_index",
                                                      torch.long, device)
    inputs["extra_args"] = {}

    return inputs

class MixSentDataFlow(DataFlow):
  def __init__(self, config, task_name, tokenizers, label_mapping):
    super().__init__(config, task_name, tokenizers, label_mapping)

  @property
  def example_class(self):
    return MixSentExample

  @property
  def minibatch_class(self):
    return MixSentMiniBatch

  def process_example(self, example: MixSentExample):
    example.process(tokenizers=self.tokenizers)

  def convert_examples_to_features(self, examples: List[MixSentExample]):
    examples: List[MixSentExample]
    features = []

    a_max_token_length = max([example.len_a for example in examples])
    b_max_token_length = max([example.len_b for example in examples])
    c_max_token_length = max([example.len_c for example in examples])

    span_a_selected_index_max_length = max([len(example.span_a_selected_index) for example in examples])
    span_b_selected_index_max_length = max([len(example.span_b_selected_index) for example in examples])
    span_c_a_selected_index_max_length = max([len(example.span_c_a_selected_index) for example in examples])
    span_c_b_selected_index_max_length = max([len(example.span_c_b_selected_index) for example in examples])

    for idx, example in enumerate(examples):
      a_padding = [0] * (a_max_token_length - example.len_a)
      b_padding = [0] * (b_max_token_length - example.len_b)
      c_padding = [0] * (c_max_token_length - example.len_c)
      a_input_ids = example.a_input_ids + [example.padding_id] * (a_max_token_length - example.len_a)
      b_input_ids = example.b_input_ids + [example.padding_id] * (b_max_token_length - example.len_b)
      c_input_ids = example.c_input_ids + [example.padding_id] * (c_max_token_length - example.len_c)

      a_segment_ids = example.a_segment_ids + a_padding
      b_segment_ids = example.b_segment_ids + b_padding
      c_segment_ids = example.c_segment_ids + c_padding

      a_input_mask = example.a_input_mask + a_padding
      b_input_mask = example.b_input_mask + b_padding
      c_input_mask = example.c_input_mask + c_padding

      a_is_head = example.a_is_head + a_padding
      b_is_head = example.b_is_head + b_padding
      c_is_head = example.c_is_head + c_padding

      span_a_selected_index = example.span_a_selected_index + [0] * (
        span_a_selected_index_max_length - len(example.span_a_selected_index))
      span_b_selected_index = example.span_b_selected_index + [0] * (
        span_b_selected_index_max_length - len(example.span_b_selected_index))
      span_c_a_selected_index = example.span_c_a_selected_index + [0] * (
        span_c_a_selected_index_max_length - len(example.span_c_a_selected_index))
      span_c_b_selected_index = example.span_c_b_selected_index + [0] * (
        span_c_b_selected_index_max_length - len(example.span_c_b_selected_index))

      features.append(
        MixSentFeature(
          a_input_ids=a_input_ids,
          b_input_ids=b_input_ids,
          c_input_ids=c_input_ids,
          a_input_mask=a_input_mask,
          b_input_mask=b_input_mask,
          c_input_mask=c_input_mask,
          a_segment_ids=a_segment_ids,
          b_segment_ids=b_segment_ids,
          c_segment_ids=c_segment_ids,
          a_is_head=a_is_head,
          b_is_head=b_is_head,
          c_is_head=c_is_head,
          span_a_selected_index=span_a_selected_index,
          span_b_selected_index=span_b_selected_index,
          span_c_a_selected_index=span_c_a_selected_index,
          span_c_b_selected_index=span_c_b_selected_index))

    return features



from relogic.logickit.training.trainer import Trainer

import torch
import json

def compute_heads_importance(config, model_trainer: Trainer, head_mask=None):
  """This method shows how to compute:
    - head attention entropy
    - head importance scores according to http://arxiv.org/abs/1905.10650
  """
  bert_config = model_trainer.model.model.encoder.bert.config
  device = next(model_trainer.model.model.encoder.parameters()).device
  n_layers, n_heads = bert_config.num_hidden_layers, bert_config.num_attention_heads
  head_importance = torch.zeros(n_layers, n_heads).to(device)
  attn_entropy = torch.zeros(n_layers, n_heads).to(device)

  if head_mask is None:
    head_mask = torch.ones(n_layers, n_heads).to(device)
  head_mask.requires_grad_(requires_grad=True)
  assert config.output_attentions, "You need to set output_attentions as True"
  results = model_trainer.analyze_task(model_trainer.tasks[0], head_mask, head_importance, attn_entropy)
  return attn_entropy, head_importance, results


def mask_heads(config, model_trainer: Trainer):
  # _, head_importance, results = compute_heads_importance(config, model_trainer)

  bert_config = model_trainer.model.model.encoder.bert.config
  device = next(model_trainer.model.model.encoder.parameters()).device
  n_layers, n_heads = bert_config.num_hidden_layers, bert_config.num_attention_heads
  head_importance = torch.zeros(n_layers, n_heads).to(device)

  new_head_mask = torch.ones_like(head_importance)
  heads_to_mask = json.load(open(config.head_to_mask_file))
  for head_to_maks in heads_to_mask:
    new_head_mask[tuple(head_to_maks)] = 0.0
  print(new_head_mask)

  _, head_importance, results = compute_heads_importance(config, model_trainer, head_mask=new_head_mask)





from bullet import Bullet, Check, YesNo, Input
from bullet import utils, colors
from bullet.client import myInput
import json
import os
import re

def check_configuration(configuration):
  return True

class ListInput:
  def __init__(
        self,
        prompt,
        default="",
        indent=0,
        word_color=colors.foreground["default"],
        strip=True,
        pattern="",
        separator=","
  ):
    self.indent = indent
    if not prompt:
      raise ValueError("Prompt can not be empty!")
    self.default = "[{}]".format(default) if default else ""
    self.prompt = prompt
    self.word_color = word_color
    self.strip = strip
    self.pattern = pattern
    self.separator = separator

  def valid(self, ans):
    if not bool(re.match(self.pattern, ans)):
      utils.moveCursorUp(1)
      utils.forceWrite(' ' * self.indent + self.prompt + self.default)
      utils.forceWrite(' ' * len(ans))
      utils.forceWrite('\b' * len(ans))
      return False
    return True

  def launch(self):
    utils.forceWrite(' ' * self.indent + self.prompt + self.default)
    sess = myInput(word_color=self.word_color)
    if not self.pattern:
      result = sess.input()
    else:
      while True:
        result = sess.input()
        if self.valid(result):
          break
    if self.strip:
      result = result.strip()
    if result:
      return result.split(self.separator)
    else:
      return []

class ConfiguredPrompt:
  def __init__(self, configuration):
    self.configuration = configuration
    self.results = {}

  def rlaunch(self, key, depth):
    results = {}
    section_config = self.configuration[key]
    if section_config["prompt_type"] == "Check":
      ui = Check(section_config["prompt"],
                 choices=section_config["choices"],
                 check=" √",
                 margin=2,
                 check_color=colors.bright(colors.foreground["red"]),
                 check_on_switch=colors.bright(colors.foreground["red"]),
                 background_color=colors.background["black"],
                 background_on_switch=colors.background["white"],
                 word_color=colors.foreground["white"],
                 word_on_switch=colors.foreground["black"],
                 indent=depth * 2)
      choices = ui.launch()
      branching = section_config.get("branching")
      if branching is not None:
        for sub_key in choices:
          branching_key = branching.get(sub_key)
          if branching_key is not None:
            if branching_key.startswith("."):
              results[sub_key] = self.rlaunch("{}{}".format(key, branching_key), depth)
            else:
              results[sub_key] = self.rlaunch(branching_key, depth)
          else:
            raise ValueError("the key {} is not in branching {}".format(sub_key, branching.keys()))
        return results
      else:
        return results
    if section_config["prompt_type"] == "ListInput":
      ui = ListInput(section_config["prompt"],
        word_color=colors.foreground["yellow"],
        indent=depth * 2)
      results = ui.launch()
      return results
    if section_config["prompt_type"] == "Input":
      ui = Input(section_config["prompt"],
        word_color=colors.foreground["yellow"],
        indent=depth * 2)
      results = ui.launch()
      return results
    if section_config["prompt_type"] == "YesNo":
      ui = YesNo(section_config["prompt"],
        word_color=colors.foreground["yellow"],
        default=section_config["default"] if "default" in section_config else 'y',
        indent=depth * 2)
      results = ui.launch()
      return results
    if section_config["prompt_type"] == "Bullet":
      ui = Bullet(section_config["prompt"],
            choices=section_config["choices"],
            bullet=" >",
            margin=2,
            bullet_color=colors.bright(colors.foreground["cyan"]),
            background_color=colors.background["black"],
            background_on_switch=colors.background["black"],
            word_color=colors.foreground["white"],
            word_on_switch=colors.foreground["white"],
            indent=depth * 2)
      results = ui.launch()
      return results
    if section_config["prompt_type"] == "GoTo":
      for sub_key in section_config["goto"]:
        if sub_key.startswith("."):
          sub_value = self.rlaunch("{}{}".format(key, sub_key), depth)
          sub_key = sub_key[1:]
        else:
          sub_value = self.rlaunch(sub_key, depth)
        if isinstance(sub_value, bool) or sub_value:
          # If True/False or other non-empty data (! "", [], {})
          results[sub_key] = sub_value
      return results

  def launch(self):
    for section in self.configuration["config_sections"]:
      utils.cprint("We are now configuring {} section".format(section), color=colors.foreground["default"])
      self.results[section] = self.rlaunch(section, 1)
    utils.cprint("You finish all the configurations, here is your full configurations",
                 color=colors.foreground["default"])
    utils.cprint(json.dumps(self.results, indent=2), color=colors.foreground["default"])
    while True:
      ui = Input("Where are you going to save the configuration? ",
          word_color=colors.foreground["yellow"],
          default="configurations/default.json")
      path = ui.launch()
      if path:
        path_dir = os.path.dirname(path)
        if os.path.exists(path_dir):
          with open(path, 'w') as fout:
            fout.write(json.dumps(self.results, indent=2))
          break
        else:
          utils.cprint("The dir `{}` does not exist".format(path_dir))
      else:
        utils.cprint("Invalid input")
    utils.cprint("Done!")

  @classmethod
  def from_config(cls, config_file):
    with open(config_file) as fin:
      configuration = json.load(fin)
      if check_configuration(configuration):
        return cls(configuration=configuration)
      else:
        raise ValueError("Configuration Error")


file_dir = os.path.dirname(os.path.abspath(__file__))
cls = ConfiguredPrompt.from_config(os.path.join(file_dir, "..", "..", "configurations/full_configuration.json"))
cls.launch()


import abc

from relogic.logickit.scorer.word_level_scorer import WordLevelScorer
from relogic.logickit.utils.utils import get_span_labels, softmax, filter_head_prediction
import os
import json



class AccuracyScorer(WordLevelScorer):
  def __init__(self, label_mapping, ignore_list=None, dump_to_file=None):
    super(AccuracyScorer, self).__init__()
    self.label_mapping = label_mapping
    self._inv_label_mapping = {v: k for k, v in label_mapping.items()}
    self.ignore_list = ignore_list if ignore_list else []
    if dump_to_file:
      self.dump_to_file_path = os.path.join(dump_to_file["output_dir"], dump_to_file["task_name"] + "_dump.json")

  def _get_results(self):
    if self.dump_to_file_path:
      self.dump_to_file_handler = open(self.dump_to_file_path, 'w')

    correct, count = 0, 0
    for example, preds in zip(self._examples, self._preds):
      confidences = [max(softmax(token_level)) for token_level in preds.data.cpu().numpy()]
      preds = preds.argmax(-1).data.cpu().numpy()
      preds = [self._inv_label_mapping[y_pred] for y_pred in preds]
      is_head = example.is_head if hasattr(example, "is_head") else example._input_token_mask
      preds = filter_head_prediction(sentence_tags=preds, is_head=is_head)
      assert len(example.labels) == len(preds)
      for y_true, y_pred in zip(example.labels, preds):
        if y_true not in self.ignore_list:
          count += 1
          correct += 1 if y_pred == y_true else 0

      if self.dump_to_file_path:
        self.dump_to_file_handler.write(
          json.dumps({
            "tokens": example.raw_tokens,
            "labels": example.labels,
            "predicted_labels": preds}) + "\n")

    if self.dump_to_file_path:
      self.dump_to_file_handler.close()

    return [
      ("accuracy", 100.0 * correct / count),
      ("loss", self.get_loss())
    ]

class F1Score(WordLevelScorer, metaclass=abc.ABCMeta):
  def __init__(self):
    super(F1Score, self).__init__()
    self._n_correct, self._n_predicted, self._n_gold = 0, 0, 0

  def _get_results(self):
    if self._n_correct == 0:
      p, r, f1 = 0, 0, 0
    else:
      p = 100.0 * self._n_correct / self._n_predicted
      r = 100.0 * self._n_correct / self._n_gold
      f1 = 2 * p * r / (p + r)
    return [
      ("total entity", self._n_gold),
      ("find entity", self._n_predicted),
      ("correct entity", self._n_correct),
      ("precision", p),
      ("recall", r),
      ("f1", f1),
      ("loss", self.get_loss()),
    ]


class EntityLevelF1Scorer(F1Score):
  def __init__(self, label_mapping, dump_to_file=None):
    super(EntityLevelF1Scorer, self).__init__()
    # label mapping : str -> int
    self._inv_label_mapping = {v: k for k, v in label_mapping.items()}
    if dump_to_file:
      self.dump_to_file_path = os.path.join(dump_to_file["output_dir"], dump_to_file["task_name"] + "_dump.json")

  def _get_results(self):
    if self.dump_to_file_path:
      self.dump_to_file_handler = open(self.dump_to_file_path, "w")

    self._n_correct, self._n_predicted, self._n_gold = 0, 0, 0
    for example, preds in zip(self._examples, self._preds):
      preds_tags = preds.argmax(-1).data.cpu().numpy()
      confidences = [max(softmax(token_level)) for token_level in preds.data.cpu().numpy()]
      is_head = example.is_head if hasattr(example, "is_head") else example._input_token_mask
      segment_ids = example.segment_ids if hasattr(example, "segment_ids") else example._input_token_mask
      if example.labels is None:
        span_preds, pred_labels = get_span_labels(
          sentence_tags=preds_tags,
          is_head=is_head,
          segment_id=segment_ids,
          inv_label_mapping=self._inv_label_mapping)
        if self.dump_to_file_path:
          self.dump_to_file_handler.write(
            json.dumps({
              "tokens": example.raw_tokens,
              "predicted_labels": pred_labels}) + "\n")
      else:
        sent_spans, sent_labels = get_span_labels(
          sentence_tags = example.labels)
        span_preds, pred_labels = get_span_labels(
          sentence_tags=preds_tags,
          is_head = is_head,
          segment_id = segment_ids,
          inv_label_mapping = self._inv_label_mapping)
        self._n_correct += len(sent_spans & span_preds)
        self._n_gold += len(sent_spans)
        self._n_predicted += len(span_preds)
        if self.dump_to_file_path:
          if len(example.raw_tokens) != len(sent_labels) or len(example.raw_tokens) != len(pred_labels):
            print(len(example.raw_tokens), example.raw_tokens)
            print(len(sent_labels), sent_labels)
            print(len(pred_labels), pred_labels)
            exit()
          self.dump_to_file_handler.write(
            json.dumps({
              "tokens": example.raw_tokens,
              "labels": sent_labels,
              "predicted_labels": pred_labels}) + "\n")

    if self.dump_to_file_path:
      self.dump_to_file_handler.close()

    return super(EntityLevelF1Scorer, self)._get_results()





from setuptools import find_packages, setup

setup(
  name='relogic00',
  version='0.0.1',
  author='Peng Shi',
  author_email="peng_shi@outlook.com",
  description='Natural Language Processing Toolkits',
  url='https://github.com/Impavidity/relogic',
  license='MIT',
  install_requires=['boto3', 'Cython', 'spacy', 'pyjnius'],
  package_data={},
  packages=find_packages(),
)
